# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import importlib

from traits.etsconfig.api import ETSConfig

from traitsui.testing.tester.target_registry import TargetRegistry
from traitsui.testing.tester._ui_tester_registry._traitsui_ui import (
    register_traitsui_ui_solvers,
)
from traitsui.ui import UI


def get_default_registries():
    """ Creates a default registry for UITester that is toolkit specific.

    Returns
    -------
    registries : list of AbstractTargetRegistry
        The default registries containing implementations for TraitsUI editors
        that are toolkit specific.
    """
    # side-effect to determine current toolkit
    from pyface.toolkit import toolkit_object  # noqa
    if ETSConfig.toolkit == "null":
        registries = []
    else:
        toolkit = {'wx': 'wx', 'qt4': 'qt4', 'qt': 'qt4'}[ETSConfig.toolkit]
        this_package, _ = __name__.rsplit(".", 1)
        module = importlib.import_module(
            ".default_registry",
            this_package + '.' + toolkit)
        registries = module.get_default_registries()

    ui_registry = TargetRegistry()
    register_traitsui_ui_solvers(
        registry=ui_registry,
        target_class=UI,
        traitsui_ui_getter=lambda target: target,
    )
    return [ui_registry] + registries



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

Implementation of a ColorEditor demo plugin for Traits UI demo program.

This demo shows each of the four styles of the ColorEditor.

Please refer to the `ColorEditor API docs`_ for further information.

.. _ColorEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.color_editor.html#traitsui.editors.color_editor.ColorEditor
"""
# Issues related to the demo warning:
# enthought/traitsui#946


from traits.api import HasTraits

from traitsui.api import Item, Group, View, Color


# Demo class definition:
class ColorEditorDemo(HasTraits):
    """ Defines the main ColorEditor demo. """

    # Define a Color trait to view:
    color_trait = Color()

    # Items are used to define the demo display, one item per editor style:
    color_group = Group(
        Item('color_trait', style='simple', label='Simple'),
        Item('_'),
        Item('color_trait', style='custom', label='Custom'),
        Item('_'),
        Item('color_trait', style='text', label='Text'),
        Item('_'),
        Item('color_trait', style='readonly', label='ReadOnly')
    )

    # Demo view
    traits_view = View(
        color_group,
        title='ColorEditor',
        buttons=['OK'],
        resizable=True
    )


# Create the demo:
demo = ColorEditorDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Tests for TabularModel (an implementation of QAbstractTableModel)
"""

import unittest

from traits.api import HasTraits, List, Str
from traitsui.api import Item, TabularEditor, View
from traitsui.tabular_adapter import TabularAdapter

from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    is_qt,
    requires_toolkit,
    reraise_exceptions,
    ToolkitName,
)
try:
    from pyface.qt import QtCore
except ImportError:
    # The entire test case should be skipped if the current backend is not Qt
    # But if it is Qt, then re-raise
    if is_qt():
        raise


class DummyHasTraits(HasTraits):
    names = List(Str)


def get_view(adapter):
    return View(
        Item(
            "names",
            editor=TabularEditor(
                adapter=adapter,
            ),
        )
    )


@requires_toolkit([ToolkitName.qt])
class TestTabularModel(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_drop_mime_data_below_list(self):
        # Test dragging an item in the list and drop it below the last item
        obj = DummyHasTraits(names=["A", "B", "C", "D"])
        view = get_view(TabularAdapter(columns=["Name"]))
        with reraise_exceptions(), \
                create_ui(obj, dict(view=view)) as ui:
            editor, = ui.get_editors("names")

            model = editor.model
            # sanity check
            self.assertEqual(model.rowCount(None), 4)

            # drag and drop row=1 from within the table.
            # drag creates a PyMimeData object for dropMimeData to consume.
            index = model.createIndex(1, 0)
            mime_data = model.mimeData([index])

            # when
            # dropped below the list, the "parent" is invalid.
            parent = QtCore.QModelIndex()   # invalid index object
            model.dropMimeData(mime_data, QtCore.Qt.MoveAction, -1, -1, parent)

            # then
            mime_data = model.mimeData(
                [model.createIndex(i, 0) for i in range(model.rowCount(None),)]
            )
            content = mime_data.instance()
            self.assertEqual(content, ["A", "C", "D", "B"])
            self.assertEqual(obj.names, content)

    def test_drop_mime_data_within_list(self):
        # Test dragging an item in the list and drop it somewhere within the
        # list
        obj = DummyHasTraits(names=["A", "B", "C", "D"])
        view = get_view(TabularAdapter(columns=["Name"]))

        with reraise_exceptions(), \
                create_ui(obj, dict(view=view)) as ui:
            editor, = ui.get_editors("names")

            model = editor.model
            # sanity check
            self.assertEqual(model.rowCount(None), 4)

            # drag and drop from within the table.
            # drag row index 0
            index = model.createIndex(0, 0)
            mime_data = model.mimeData([index])

            # when
            # drop it to row index 2
            parent = model.createIndex(2, 0)
            model.dropMimeData(mime_data, QtCore.Qt.MoveAction, -1, -1, parent)

            # then
            mime_data = model.mimeData(
                [model.createIndex(i, 0) for i in range(model.rowCount(None),)]
            )
            content = mime_data.instance()
            self.assertEqual(content, ["B", "C", "A", "D"])
            self.assertEqual(obj.names, content)

    def test_copy_item(self):
        # Test copy 'A' to the row after 'C'
        obj = DummyHasTraits(names=["A", "B", "C"])
        view = get_view(TabularAdapter(columns=["Name"], can_drop=True))

        with reraise_exceptions(), \
                create_ui(obj, dict(view=view)) as ui:
            editor, = ui.get_editors("names")

            model = editor.model
            # sanity check
            self.assertEqual(model.rowCount(None), 3)

            # drag and drop from within the table for copy action.
            # drag index 0
            index = model.createIndex(0, 0)
            mime_data = model.mimeData([index])

            # when
            # drop to index 2
            parent = model.createIndex(2, 0)
            model.dropMimeData(mime_data, QtCore.Qt.CopyAction, -1, -1, parent)

            # then
            self.assertEqual(model.rowCount(None), 4)
            mime_data = model.mimeData(
                [model.createIndex(i, 0) for i in range(model.rowCount(None),)]
            )
            content = mime_data.instance()
            self.assertEqual(content, ["A", "B", "C", "A"])
            self.assertEqual(obj.names, content)

    def test_move_rows_invalid_index(self):
        # Test the last resort to prevent segfault

        obj = DummyHasTraits(names=["A", "B", "C"])
        view = get_view(TabularAdapter(columns=["Name"]))

        with reraise_exceptions(), \
                create_ui(obj, dict(view=view)) as ui:
            editor, = ui.get_editors("names")

            model = editor.model
            # sanity check
            self.assertEqual(model.rowCount(None), 3)

            # when
            # -1 is an invalid row. This should not cause segfault.
            model.moveRows([1], -1)

            # then
            mime_data = model.mimeData(
                [model.createIndex(i, 0) for i in range(model.rowCount(None),)]
            )
            content = mime_data.instance()
            self.assertEqual(content, ["A", "C", "B"])
            self.assertEqual(obj.names, content)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Test case for bug (wx, Mac OS X)

A RangeEditor in mode 'text' for an Int allows values out of range.
"""
import unittest

from traits.has_traits import HasTraits
from traits.trait_types import Float, Int
from traitsui.item import Item
from traitsui.view import View
from traitsui.editors.range_editor import RangeEditor

from traitsui.testing.tester import command
from traitsui.testing.tester.ui_tester import UITester
from traitsui.tests._tools import (
    BaseTestMixin,
    requires_toolkit,
    ToolkitName,
)


class NumberWithRangeEditor(HasTraits):
    """Dialog containing a RangeEditor in 'spinner' mode for an Int.
    """

    number = Int()

    traits_view = View(
        Item(label="Range should be 3 to 8. Enter 1, then press OK"),
        Item("number", editor=RangeEditor(low=3, high=8, mode="text")),
        buttons=["OK"],
    )


class FloatWithRangeEditor(HasTraits):
    """Dialog containing a RangeEditor in 'spinner' mode for an Int.
    """

    number = Float(5.0)

    traits_view = View(
        Item("number", editor=RangeEditor(low=0.0, high=12.0)), buttons=["OK"]
    )


class TestRangeEditorText(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @requires_toolkit([ToolkitName.wx])
    def test_wx_text_editing(self):
        # behavior: when editing the text part of a spin control box, pressing
        # the OK button should update the value of the HasTraits class
        # (tests a bug where this fails with an AttributeError)

        num = NumberWithRangeEditor()
        tester = UITester()
        with tester.create_ui(num) as ui:
            # the following is equivalent to setting the text in the text
            # control, then pressing OK
            text = tester.find_by_name(ui, "number")
            text.perform(command.KeyClick("1"))
            text.perform(command.KeyClick("Enter"))

        # the number traits should be between 3 and 8
        self.assertTrue(3 <= num.number <= 8)


if __name__ == "__main__":
    # Executing the file opens the dialog for manual testing
    num = NumberWithRangeEditor()
    num.configure_traits()
    print(num.number)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Test FontEditor
"""
import unittest

from traits.api import HasTraits
from traitsui.api import Font, Item, View
from traitsui.tests._tools import (
    requires_toolkit,
    ToolkitName,
)
from traitsui.testing.api import (
    MouseClick,
    UITester
)


class ObjectWithFont(HasTraits):
    font_trait = Font()


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestFontEditor(unittest.TestCase):

    def test_create_and_dispose_text_style(self):
        # Setting focus on the widget and then disposing the widget
        # should not cause errors.
        view = View(Item("font_trait", style="text"))
        tester = UITester()
        with tester.create_ui(ObjectWithFont(), dict(view=view)) as ui:
            wrapper = tester.find_by_name(ui, "font_trait")
            wrapper.perform(MouseClick())



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the various font editors and the font editor factory, for the
    wxPython user interface toolkit..
"""


import wx

from traits.api import Bool

from traitsui.editors.font_editor import (
    ToolkitEditorFactory as BaseToolkitEditorFactory,
)

from .editor_factory import (
    SimpleEditor as BaseSimpleEditor,
    TextEditor as BaseTextEditor,
    ReadonlyEditor as BaseReadonlyEditor,
)

from .editor import Editor

from .helper import TraitsUIPanel, disconnect


# Standard font point sizes
PointSizes = [
    "8",
    "9",
    "10",
    "11",
    "12",
    "14",
    "16",
    "18",
    "20",
    "22",
    "24",
    "26",
    "28",
    "36",
    "48",
    "72",
]

# All available font styles
Styles = ["Normal", "Slant", "Italic"]

# All available font weights
Weights = ["Normal", "Light", "Bold"]

# All available font facenames
facenames = None

# ---------------------------------------------------------------------------
#  The wxPython ToolkitEditorFactory class.
# ---------------------------------------------------------------------------
## We need to add wx-specific methods to the editor factory, and so we create
## a subclass of the BaseToolkitEditorFactory.


class ToolkitEditorFactory(BaseToolkitEditorFactory):
    """ wxPython editor factory for font editors.
    """

    show_style = Bool(False)
    show_weight = Bool(False)

    def to_wx_font(self, editor):
        """ Returns a wxFont object corresponding to a specified object's font
        trait.
        """
        font = editor.value
        return wx.Font(
            font.GetPointSize(),
            font.GetFamily(),
            font.GetStyle(),
            font.GetWeight(),
            font.GetUnderlined(),
            font.GetFaceName(),
        )

    def from_wx_font(self, font):
        """ Gets the application equivalent of a wxPython Font value.
        """
        return font

    def str_font(self, font):
        """ Returns the text representation of the specified object trait value.
        """
        weight = {
            wx.FONTWEIGHT_LIGHT: " Light",
            wx.FONTWEIGHT_BOLD: " Bold",
        }.get(font.GetWeight(), "")
        style = {
            wx.FONTSTYLE_SLANT: " Slant",
            wx.FONTSTYLE_ITALIC: " Italic",
        }.get(font.GetStyle(), "")
        return "%s point %s%s%s" % (
            font.GetPointSize(),
            font.GetFaceName(),
            style,
            weight,
        )

    def all_facenames(self):
        """ Returns a list of all available font facenames.
        """
        global facenames

        if facenames is None:
            facenames = sorted(FontEnumerator().facenames())
        return facenames


# -------------------------------------------------------------------------
#  'SimpleFontEditor' class:
# -------------------------------------------------------------------------


class SimpleFontEditor(BaseSimpleEditor):
    """ Simple style of font editor, which displays a text field that contains
        a text representation of the font value (using that font if possible).
        Clicking the field displays a font selection dialog box.
    """

    def popup_editor(self, event):
        """ Invokes the pop-up editor for an object trait.
        """
        font_data = wx.FontData()
        font_data.SetInitialFont(self.factory.to_wx_font(self))
        dialog = wx.FontDialog(self.control, font_data)
        if dialog.ShowModal() == wx.ID_OK:
            self.value = self.factory.from_wx_font(
                dialog.GetFontData().GetChosenFont()
            )
            self.update_editor()

        dialog.Destroy()

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        super().update_editor()
        set_font(self)

    def string_value(self, font):
        """ Returns the text representation of a specified font value.
        """
        return self.factory.str_font(font)


class CustomFontEditor(Editor):
    """ Custom style of font editor, which displays the following:

        * A combo box containing all available type face names.
        * A combo box containing the available type sizes.
        * A combo box containing the available type styles
    """

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        # Create a panel to hold all of the buttons:
        self.control = panel = TraitsUIPanel(parent, -1)
        sizer = wx.BoxSizer(wx.VERTICAL)

        # Add all of the font choice controls:
        sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        facenames = self.factory.all_facenames()
        control = self._facename = wx.Choice(
            panel, -1, wx.Point(0, 0), wx.Size(-1, -1), facenames
        )

        sizer2.Add(control, 4, wx.EXPAND)
        panel.Bind(wx.EVT_CHOICE, self.update_object_parts, id=control.GetId())

        control = self._point_size = wx.Choice(
            panel, -1, wx.Point(0, 0), wx.Size(-1, -1), PointSizes
        )
        sizer2.Add(control, 1, wx.EXPAND | wx.LEFT, 3)
        panel.Bind(wx.EVT_CHOICE, self.update_object_parts, id=control.GetId())

        if self.factory.show_style:
            self._style = wx.Choice(
                panel, -1, wx.Point(0, 0), wx.Size(-1, -1), Styles
            )
            sizer2.Add(self._style, 1, wx.EXPAND | wx.LEFT, 3)
            panel.Bind(wx.EVT_CHOICE, self.update_object_parts, id=self._style.GetId())

        if self.factory.show_weight:
            self._weight = wx.Choice(
                panel, -1, wx.Point(0, 0), wx.Size(-1, -1), Weights
            )
            sizer2.Add(self._weight, 1, wx.EXPAND | wx.LEFT, 3)
            panel.Bind(wx.EVT_CHOICE, self.update_object_parts, id=self._weight.GetId())

        sizer.Add(sizer2, 0, wx.EXPAND)

        # Set-up the layout:
        panel.SetSizer(sizer)

        self.set_tooltip()

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        disconnect(self._facename, wx.EVT_CHOICE)
        disconnect(self._point_size, wx.EVT_CHOICE)
        if self.factory.show_style:
            disconnect(self._style, wx.EVT_CHOICE)
        if self.factory.show_weight:
            disconnect(self._weight, wx.EVT_CHOICE)

        super().dispose()

    def update_object_parts(self, event):
        """ Handles the user modifying one of the font components.
        """
        point_size = int(self._point_size.GetStringSelection())
        facename = self._facename.GetStringSelection()

        style = wx.FONTSTYLE_NORMAL
        if self.factory.show_style:
            style += self._style.GetCurrentSelection()

        weight = wx.FONTWEIGHT_NORMAL
        if self.factory.show_weight:
            weight += self._weight.GetCurrentSelection()

        font = wx.Font(
            point_size, wx.FONTFAMILY_DEFAULT, style, weight, faceName=facename
        )
        self.value = self.factory.from_wx_font(font)

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        font = self.factory.to_wx_font(self)

        try:
            self._facename.SetStringSelection(font.GetFaceName())
        except:
            self._facename.SetSelection(0)

        try:
            self._point_size.SetStringSelection(str(font.GetPointSize()))
        except:
            self._point_size.SetSelection(0)

    def string_value(self, font):
        """ Returns the text representation of a specified font value.
        """
        return self.factory.str_font(font)


# -------------------------------------------------------------------------
#  'TextFontEditor' class:
# -------------------------------------------------------------------------


class TextFontEditor(BaseTextEditor):
    """ Text style of font editor, which displays an editable text field
        containing a text representation of the font value (using that font if
        possible).
    """

    def update_object(self, event):
        """ Handles the user changing the contents of the edit control.
        """
        self.value = self.control.GetValue()

    def update_editor(self):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        super().update_editor()
        set_font(self)

    def string_value(self, font):
        """ Returns the text representation of a specified font value.
        """
        return self.factory.str_font(font)


class ReadonlyFontEditor(BaseReadonlyEditor):
    """ Read-only style of font editor, which displays a read-only text field
        containing a text representation of the font value (using that font if
        possible).
    """

    def update_editor(self):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        super().update_editor()
        set_font(self)

    def string_value(self, font):
        """ Returns the text representation of a specified font value.
        """
        return self.factory.str_font(font)


# -------------------------------------------------------------------------
#  Set the editor control's font to match a specified font:
# -------------------------------------------------------------------------


def set_font(editor):
    """ Sets the editor control's font to match a specified font.
    """
    font = editor.factory.to_wx_font(editor)
    font.SetPointSize(min(10, font.GetPointSize()))
    editor.control.SetFont(font)


class FontEnumerator(wx.FontEnumerator):
    """ An enumeration of fonts.
    """

    def facenames(self):
        """ Returns a list of all available font facenames.
        """
        self._facenames = []
        self.EnumerateFacenames()
        return self._facenames

    def OnFacename(self, facename):
        """ Adds a facename to the list of facenames.
        """
        self._facenames.append(facename)
        return True


# Define the names SimpleEditor, CustomEditor, TextEditor and ReadonlyEditor
# which are looked up by the editor factory for the font editor.
SimpleEditor = SimpleFontEditor
CustomEditor = CustomFontEditor
TextEditor = TextFontEditor
ReadonlyEditor = ReadonlyFontEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the code editor factory for all traits toolkit backends,
useful for tools such as debuggers.
"""

from traits.api import Instance, Str, Enum, Bool

from traitsui.editor_factory import EditorFactory
from traitsui.toolkit_traits import Color


class CodeEditor(EditorFactory):
    """ Editor factory for code editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Object trait containing list of line numbers to mark (optional)
    mark_lines = Str()

    #: Background color for marking lines
    mark_color = Color(0xECE9D8)

    #: Object trait containing the currently selected line (optional)
    selected_line = Str()

    #: Object trait containing the currently selected text (optional)
    selected_text = Str()

    #: Object trait containing the currently selected text start position
    #: (optional)
    selected_start_pos = Str()

    #: Object trait containing the currently selected text end position
    #: (optional)
    selected_end_pos = Str()

    #: Background color for selected lines
    selected_color = Color(0xA4FFFF)

    #: Where should the search toolbar be placed?
    search = Enum("top", "bottom", "none")

    #: Background color for lines that match the current search
    search_color = Color(0xFFFF94)

    #: Current line
    line = Str()

    #: Current column
    column = Str()

    #: Should code folding be enabled?
    foldable = Bool(True)

    #: Should line numbers be displayed in the margin?
    show_line_numbers = Bool(True)

    #: Is user input set on every change?
    auto_set = Bool(True)

    #: Should the editor auto-scroll when a new **selected_line** value is set?
    auto_scroll = Bool(True)

    #: Optional key bindings associated with the editor
    key_bindings = Instance("traitsui.key_bindings.KeyBindings")

    #: Calltip clicked event
    calltip_clicked = Str()

    #: The lexer to use. Default is 'python'; 'null' indicates no lexing.
    lexer = Str("python")

    #: Object trait containing the list of line numbers to dim (optional)
    dim_lines = Str()

    #: Object trait to dim lines to. Can be of form #rrggbb or a color spec. If
    #: not specified, dark grey is used.
    dim_color = Str()

    #: Object trait containing the list of line numbers to put squiggles under
    #: (optional)
    squiggle_lines = Str()

    #: Object trait for the color of squiggles. If not specified, red is used.
    squiggle_color = Str()


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = CodeEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the compound editor and the compound editor factory for the
    wxPython user interface toolkit.
"""


import wx

from traits.api import Str

from .editor import Editor

from .helper import TraitsUIPanel

# -------------------------------------------------------------------------
#  'CompoundEditor' class:
# -------------------------------------------------------------------------


class CompoundEditor(Editor):
    """ Editor for compound traits, which displays editors for each of the
    combined traits, in the appropriate style.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The kind of editor to create for each list item
    kind = Str()

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        # Create a panel to hold all of the component trait editors:
        self.control = panel = TraitsUIPanel(parent, -1)
        sizer = wx.BoxSizer(wx.VERTICAL)

        # Add all of the component trait editors:
        self._editors = editors = []
        for factory in self.factory.editors:
            editor = getattr(factory, self.kind)(
                self.ui, self.object, self.name, self.description, panel
            )
            editor.prepare(panel)
            sizer.Add(
                editor.control, 1, wx.TOP | wx.BOTTOM | editor.layout_style, 3
            )
            editors.append(editor)

        # Set-up the layout:
        panel.SetSizerAndFit(sizer)

        self.set_tooltip()

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        pass

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        for editor in self._editors:
            editor.dispose()

        super().dispose()


class SimpleEditor(CompoundEditor):

    #: The kind of editor to create for each list item. This value overrides
    #: the default.
    kind = "simple_editor"


class CustomEditor(CompoundEditor):

    #: The kind of editor to create for each list item. This value overrides
    #: the default.

    kind = "custom_editor"



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines a text editor which displays a text field and maintains a history
    of previously entered values.
"""

from traits.api import Int, Bool

from traitsui.basic_editor_factory import BasicEditorFactory
from traitsui.toolkit import toolkit_object

# Define callable which returns the 'klass' value (i.e. the editor to use in
# the EditorFactory.


def history_editor(*args, **traits):
    return toolkit_object("history_editor:_HistoryEditor")(*args, **traits)


# -------------------------------------------------------------------------
#  Create the editor factory object:
# -------------------------------------------------------------------------


class ToolkitEditorFactory(BasicEditorFactory):

    #: The number of entries in the history:
    entries = Int(10)

    #: Should each keystroke update the value (or only the enter key, tab,
    #: etc.)?
    auto_set = Bool(False)


HistoryEditor = ToolkitEditorFactory(klass=history_editor)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traitsui.testing.tester.target_registry import (
    TargetRegistry,
)
from traitsui.testing.tester.exceptions import (
    InteractionNotSupported,
    LocationNotSupported,
)


class TestInteractionRegistry(unittest.TestCase):

    def test_registry_empty(self):

        class SpecificTarget:
            pass

        class Action:
            pass

        registry = TargetRegistry()
        with self.assertRaises(InteractionNotSupported) as exception_context:
            registry._get_handler(SpecificTarget(), Action())

        self.assertEqual(
            str(exception_context.exception),
            f"No handler is found for target {SpecificTarget!r} with "
            f"interaction {Action!r}. Supported these: []",
        )

    def test_register_editor_with_action(self):
        registry = TargetRegistry()

        class SpecificEditor:
            pass

        class UserAction:
            pass

        def handler(wrapper, interaction):
            pass

        # when
        registry.register_interaction(
            target_class=SpecificEditor,
            interaction_class=UserAction,
            handler=handler,
        )

        # then
        actual = registry._get_handler(SpecificEditor(), UserAction())
        self.assertIs(actual, handler)

    def test_get_interactions_supported(self):
        registry = TargetRegistry()

        class SpecificEditor:
            pass

        class UserAction:
            pass

        class UserAction2:
            pass

        def handler(wrapper, interaction):
            pass

        # when
        registry.register_interaction(
            target_class=SpecificEditor,
            interaction_class=UserAction,
            handler=handler,
        )
        registry.register_interaction(
            target_class=SpecificEditor,
            interaction_class=UserAction2,
            handler=handler,
        )

        # then
        self.assertEqual(
            registry._get_interactions(SpecificEditor()),
            {UserAction, UserAction2}
        )

    def test_action_not_supported_report_supported_action(self):
        # Test raise InteractionNotSupported contains information about what
        # actions are supported.

        class SpecificEditor:
            pass

        class SpecificEditor2:
            pass

        class UserAction:
            pass

        class UserAction2:
            pass

        class UserAction3:
            pass

        def handler(wrapper, interaction):
            pass

        registry = TargetRegistry()
        registry.register_interaction(SpecificEditor, UserAction, handler)
        registry.register_interaction(SpecificEditor2, UserAction2, handler)
        registry.register_interaction(SpecificEditor2, UserAction3, handler)

        with self.assertRaises(InteractionNotSupported) as exception_context:
            registry._get_handler(SpecificEditor2(), None)

        self.assertIn(UserAction2, exception_context.exception.supported)
        self.assertIn(UserAction3, exception_context.exception.supported)
        self.assertNotIn(UserAction, exception_context.exception.supported)

    def test_error_conflict(self):
        # Test the same target + interaction type cannot be registered twice.

        class SpecificEditor:
            pass

        class UserAction:
            pass

        def handler(wrapper, interaction):
            pass

        registry = TargetRegistry()
        registry.register_interaction(SpecificEditor, UserAction, handler)

        with self.assertRaises(ValueError):
            registry.register_interaction(SpecificEditor, UserAction, handler)

    def test_error_get_interaction_doc(self):
        # The registry is empty
        registry = TargetRegistry()
        with self.assertRaises(InteractionNotSupported):
            registry._get_interaction_doc(2.1, int)

    def test_get_default_interaction_doc(self):

        class Action:
            """Some action."""
            pass

        def handler(wrapper, interaction):
            pass

        registry = TargetRegistry()
        registry.register_interaction(
            target_class=float,
            interaction_class=Action,
            handler=handler,
        )

        actual = registry._get_interaction_doc(
            target=21.2, interaction_class=Action,
        )
        self.assertEqual(actual, "Some action.")


class TestLocationRegistry(unittest.TestCase):

    def test_location_registry_empty(self):

        class SpecificTarget:
            pass

        class Locator:
            pass

        registry = TargetRegistry()
        with self.assertRaises(LocationNotSupported) as exception_context:
            registry._get_solver(SpecificTarget(), Locator())

        self.assertEqual(exception_context.exception.supported, [])
        self.assertEqual(
            str(exception_context.exception),
            f"Location {Locator!r} is not supported for {SpecificTarget!r}. "
            "Supported these: []",
        )

    def test_register_location(self):

        def solver(wrapper, location):
            return 1

        registry = TargetRegistry()
        registry.register_location(
            target_class=float, locator_class=str, solver=solver)

        self.assertIs(registry._get_solver(2.1, "dummy"), solver)

    def test_register_location_report_existing(self):

        def solver(wrapper, location):
            return 1

        registry = TargetRegistry()
        registry.register_location(
            target_class=float, locator_class=str, solver=solver)

        with self.assertRaises(LocationNotSupported) as exception_context:
            registry._get_solver(3.4, None)

        self.assertEqual(exception_context.exception.supported, [str])

    def test_get_locations_supported(self):
        # Test _get_locations return the supported location types.
        registry = TargetRegistry()

        class SpecificEditor:
            pass

        class Locator1:
            pass

        class Locator2:
            pass

        def solver(wrapper, location):
            return 1

        # when
        registry.register_location(
            target_class=SpecificEditor,
            locator_class=Locator1,
            solver=solver,
        )
        registry.register_location(
            target_class=SpecificEditor,
            locator_class=Locator2,
            solver=solver,
        )

        # then
        self.assertEqual(
            registry._get_locations(SpecificEditor()),
            {Locator1, Locator2}
        )

    def test_get_location_help_default(self):

        class Locator:
            """ Some default documentation."""
            pass

        registry = TargetRegistry()
        registry.register_location(
            target_class=float,
            locator_class=Locator,
            solver=lambda w, l: 1,
        )

        help_text = registry._get_location_doc(
            target=2.345, locator_class=Locator,
        )
        self.assertEqual(help_text, "Some default documentation.")

    def test_error_get_interaction_doc(self):
        # The registry is empty
        registry = TargetRegistry()
        with self.assertRaises(LocationNotSupported):
            registry._get_location_doc(3.456, int)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# default_traits_view2.py -- Sample code to demonstrate the use of
# 'default_traits_view'

from traits.api import HasTraits, Str, Int
from traitsui.api import View, Item, Group


class SimpleEmployee2(HasTraits):
    first_name = Str()
    last_name = Str()
    department = Str()

    employee_number = Str()
    salary = Int()

    def default_traits_view(self):
        return View(
            Group(
                Item(name='first_name'),
                Item(name='last_name'),
                Item(name='department'),
                label='Personnel profile',
                show_border=True,
            ),
        )


sam = SimpleEmployee2()
sam.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the table column descriptor used by the editor and editor factory
    classes for numeric and table editors.
"""

import os

from traits.api import (
    Any,
    Bool,
    Callable,
    Constant,
    Enum,
    Expression,
    Float,
    HasPrivateTraits,
    Instance,
    Int,
    Property,
    Str,
    Either,
)

from traits.trait_base import user_name_for, xgetattr

from .editor_factory import EditorFactory
from .menu import Menu
from .ui_traits import Image, AView, EditorStyle
from .toolkit_traits import Color, Font
from .view import View

# Set up a logger:
import logging


logger = logging.getLogger(__name__)


# Flag used to indicate user has not specified a column label
UndefinedLabel = "???"


class TableColumn(HasPrivateTraits):
    """ Represents a column in a table editor.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Column label to use for this column:
    label = Str(UndefinedLabel)

    #: Type of data contained by the column:
    # XXX currently no other types supported, but potentially there could be...
    type = Enum("text", "bool")

    #: Text color for this column:
    text_color = Color("black")

    #: Text font for this column:
    text_font = Either(None, Font)

    #: Cell background color for this column:
    cell_color = Color("white", allow_none=True)

    #: Cell background color for non-editable columns:
    read_only_cell_color = Color(0xF4F3EE, allow_none=True)

    #: Cell graph color:
    graph_color = Color(0xDDD9CC)

    #: Horizontal alignment of text in the column:
    horizontal_alignment = Enum("left", ["left", "center", "right"])

    #: Vertical alignment of text in the column:
    vertical_alignment = Enum("center", ["top", "center", "bottom"])

    #: Horizontal cell margin
    horizontal_margin = Int(4)

    #: Vertical cell margin
    vertical_margin = Int(3)

    #: The image to display in the cell:
    image = Image

    #: Renderer used to render the contents of this column:
    renderer = Any  # A toolkit specific renderer

    #: Is the table column visible (i.e., viewable)?
    visible = Bool(True)

    #: Is this column editable?
    editable = Bool(True)

    #: Is the column automatically edited/viewed (i.e. should the column editor
    #: or popup be activated automatically on mouse over)?
    auto_editable = Bool(False)

    #: Should a checkbox be displayed instead of True/False?
    show_checkbox = Bool(True)

    #: Can external objects be dropped on the column?
    droppable = Bool(False)

    #: Context menu to display when this column is right-clicked:
    menu = Instance(Menu)

    #: The tooltip to display when the mouse is over the column:
    tooltip = Str()

    #: The width of the column (< 0.0: Default, 0.0..1.0: fraction of total
    #: table width, > 1.0: absolute width in pixels):
    width = Float(-1.0)

    #: The width of the column while it is being edited (< 0.0: Default,
    #: 0.0..1.0: fraction of total table width, > 1.0: absolute width in
    #: pixels):
    edit_width = Float(-1.0)

    #: The height of the column cell's row while it is being edited
    #: (< 0.0: Default, 0.0..1.0: fraction of total table height,
    #: > 1.0: absolute height in pixels):
    edit_height = Float(-1.0)

    #: The resize mode for this column.  This takes precedence over other
    #: settings (like **width**, above).
    #: - "interactive": column can be resized by users or programmatically
    #: - "fixed": users cannot resize the column, but it can be set programmatically
    #: - "stretch": the column will be resized to fill the available space
    #: - "resize_to_contents": column will be sized to fit the contents, but then cannot be resized
    resize_mode = Enum("interactive", "fixed", "stretch", "resize_to_contents")

    #: The view (if any) to display when clicking a non-editable cell:
    view = AView

    #: Optional maximum value a numeric cell value can have:
    maximum = Float(trait_value=True)

    # -------------------------------------------------------------------------
    #:  Returns the actual object being edited:
    # -------------------------------------------------------------------------

    def get_object(self, object):
        """ Returns the actual object being edited.
        """
        return object

    def get_label(self):
        """ Gets the label of the column.
        """
        return self.label

    def get_width(self):
        """ Returns the width of the column.
        """
        return self.width

    def get_edit_width(self, object):
        """ Returns the edit width of the column.
        """
        return self.edit_width

    def get_edit_height(self, object):
        """ Returns the height of the column cell's row while it is being
            edited.
        """
        return self.edit_height

    def get_type(self, object):
        """ Gets the type of data for the column for a specified object.
        """
        return self.type

    def get_text_color(self, object):
        """ Returns the text color for the column for a specified object.
        """
        return self.text_color_

    def get_text_font(self, object):
        """ Returns the text font for the column for a specified object.
        """
        return self.text_font

    def get_cell_color(self, object):
        """ Returns the cell background color for the column for a specified
            object.
        """
        if self.is_editable(object):
            return self.cell_color_
        return self.read_only_cell_color_

    def get_graph_color(self, object):
        """ Returns the cell background graph color for the column for a
            specified object.
        """
        return self.graph_color_

    def get_horizontal_alignment(self, object):
        """ Returns the horizontal alignment for the column for a specified
            object.
        """
        return self.horizontal_alignment

    def get_vertical_alignment(self, object):
        """ Returns the vertical alignment for the column for a specified
            object.
        """
        return self.vertical_alignment

    def get_image(self, object):
        """ Returns the image to display for the column for a specified object.
        """
        return self.image

    def get_renderer(self, object):
        """ Returns the renderer for the column of a specified object.
        """
        return self.renderer

    def is_editable(self, object):
        """ Returns whether the column is editable for a specified object.
        """
        return self.editable

    def is_auto_editable(self, object):
        """ Returns whether the column is automatically edited/viewed for a
            specified object.
        """
        return self.auto_editable

    def is_droppable(self, object, value):
        """ Returns whether a specified value is valid for dropping on the
            column for a specified object.
        """
        return self.droppable

    def get_menu(self, object):
        """ Returns the context menu to display when the user right-clicks on
            the column for a specified object.
        """
        return self.menu

    def get_tooltip(self, object):
        """ Returns the tooltip to display when the user mouses over the column
            for a specified object.
        """
        return self.tooltip

    def get_view(self, object):
        """ Returns the view to display when clicking a non-editable cell.
        """
        return self.view

    def get_maximum(self, object):
        """ Returns the maximum value a numeric column can have.
        """
        return self.maximum

    def on_click(self, object):
        """ Called when the user clicks on the column.
        """
        pass

    def on_dclick(self, object):
        """ Called when the user clicks on the column.
        """
        pass

    def cmp(self, object1, object2):
        """ Returns the result of comparing the column of two different objects.

        This is deprecated.
        """
        return (self.key(object1) > self.key(object2)) - (
            self.key(object1) < self.key(object2)
        )

    def __str__(self):
        """ Returns the string representation of the table column.
        """
        return self.get_label()


class ObjectColumn(TableColumn):
    """ A column for editing objects.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Name of the object trait associated with this column:
    name = Str()

    #: Column label to use for this column:
    label = Property()

    #: Trait editor used to edit the contents of this column:
    editor = Instance(EditorFactory)

    #: The editor style to use to edit the contents of this column:
    style = EditorStyle

    #: Format string to apply to column values:
    format = Str("%s")

    #: Format function to apply to column values:
    format_func = Callable()

    # -------------------------------------------------------------------------
    #  Trait view definitions:
    # -------------------------------------------------------------------------

    traits_view = View(
        [
            ["name", "label", "type", "|[Column Information]"],
            [
                "horizontal_alignment{Horizontal}@",
                "vertical_alignment{Vertical}@",
                "|[Alignment]",
            ],
            ["editable", "9", "droppable", "9", "visible", "-[Options]>"],
            "|{Column}",
        ],
        [
            [
                "text_color@",
                "cell_color@",
                "read_only_cell_color@",
                "|[UI Colors]",
            ],
            "|{Colors}",
        ],
        [["text_font@", "|[Font]<>"], "|{Font}"],
        ["menu@", "|{Menu}"],
        ["editor@", "|{Editor}"],
    )

    def _get_label(self):
        """ Gets the label of the column.
        """
        if self._label is not None:
            return self._label
        return user_name_for(self.name)

    def _set_label(self, label):
        old, self._label = self._label, label
        if old != label:
            self.trait_property_changed("label", old, label)

    def get_raw_value(self, object):
        """ Gets the unformatted value of the column for a specified object.
        """
        try:
            return xgetattr(self.get_object(object), self.name)
        except Exception as e:
            from traitsui.api import raise_to_debug

            raise_to_debug()
            return None

    def get_value(self, object):
        """ Gets the formatted value of the column for a specified object.
        """
        try:
            if self.format_func is not None:
                return self.format_func(self.get_raw_value(object))

            return self.format % (self.get_raw_value(object),)
        except:
            logger.exception(
                "Error occurred trying to format a %s value"
                % self.__class__.__name__
            )
            return "Format!"

    def get_drag_value(self, object):
        """Returns the drag value for the column.
        """
        return self.get_raw_value(object)

    def set_value(self, object, value):
        """ Sets the value of the column for a specified object.
        """
        target, name = self.target_name(object)
        setattr(target, name, value)

    def get_editor(self, object):
        """ Gets the editor for the column of a specified object.
        """
        if self.editor is not None:
            return self.editor

        target, name = self.target_name(object)

        return target.base_trait(name).get_editor()

    def get_style(self, object):
        """ Gets the editor style for the column of a specified object.
        """
        return self.style

    def key(self, object):
        """ Returns the value to use for sorting.
        """
        return self.get_raw_value(object)

    def is_droppable(self, object, value):
        """ Returns whether a specified value is valid for dropping on the
            column for a specified object.
        """
        if self.droppable:
            try:
                target, name = self.target_name(object)
                target.base_trait(name).validate(target, name, value)
                return True
            except:
                pass

        return False

    def target_name(self, object):
        """ Returns the target object and name for the column.
        """
        object = self.get_object(object)
        name = self.name
        col = name.rfind(".")
        if col < 0:
            return (object, name)

        return (xgetattr(object, name[:col]), name[col + 1 :])


class ExpressionColumn(ObjectColumn):
    """ A column for displaying computed values.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The Python expression used to return the value of the column:
    expression = Expression

    #: Is this column editable?
    editable = Constant(False)

    #: The globals dictionary that should be passed to the expression
    #: evaluation:
    globals = Any({})

    def get_raw_value(self, object):
        """ Gets the unformatted value of the column for a specified object.
        """
        try:
            return eval(self.expression_, self.globals, {"object": object})
        except Exception:
            logger.exception(
                "Error evaluating table column expression: %s"
                % self.expression
            )
            return None


class NumericColumn(ObjectColumn):
    """ A column for editing Numeric arrays.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Column label to use for this column
    label = Property()

    #: Text color this column when selected
    selected_text_color = Color("black")

    #: Text font for this column when selected
    selected_text_font = Font

    #: Cell background color for this column when selected
    selected_cell_color = Color(0xD8FFD8)

    #: Formatting string for the cell value
    format = Str("%s")

    #: Horizontal alignment of text in the column; this value overrides the
    #: default.
    horizontal_alignment = "center"

    def _get_label(self):
        """ Gets the label of the column.
        """
        if self._label is not None:
            return self._label
        return self.name

    def _set_label(self, label):
        old, self._label = self._label, label
        if old != label:
            self.trait_property_changed("label", old, label)

    def get_type(self, object):
        """ Gets the type of data for the column for a specified object row.
        """
        return self.type

    def get_text_color(self, object):
        """ Returns the text color for the column for a specified object row.
        """
        if self._is_selected(object):
            return self.selected_text_color_
        return self.text_color_

    def get_text_font(self, object):
        """ Returns the text font for the column for a specified object row.
        """
        if self._is_selected(object):
            return self.selected_text_font
        return self.text_font

    def get_cell_color(self, object):
        """ Returns the cell background color for the column for a specified
            object row.
        """
        if self.is_editable(object):
            if self._is_selected(object):
                return self.selected_cell_color_
            return self.cell_color_
        return self.read_only_cell_color_

    def get_horizontal_alignment(self, object):
        """ Returns the horizontal alignment for the column for a specified
            object row.
        """
        return self.horizontal_alignment

    def get_vertical_alignment(self, object):
        """ Returns the vertical alignment for the column for a specified
            object row.
        """
        return self.vertical_alignment

    def is_editable(self, object):
        """ Returns whether the column is editable for a specified object row.
        """
        return self.editable

    def is_droppable(self, object, row, value):
        """ Returns whether a specified value is valid for dropping on the
            column for a specified object row.
        """
        return self.droppable

    def get_menu(self, object, row):
        """ Returns the context menu to display when the user right-clicks on
            the column for a specified object row.
        """
        return self.menu

    def get_value(self, object):
        """ Gets the value of the column for a specified object row.
        """
        try:
            value = getattr(object, self.name)
            try:
                return self.format % (value,)
            except:
                return "Format!"
        except:
            return "Undefined!"

    def set_value(self, object, row, value):
        """ Sets the value of the column for a specified object row.
        """
        column = self.get_data_column(object)
        column[row] = type(column[row])(value)

    def get_editor(self, object):
        """ Gets the editor for the column of a specified object row.
        """
        return super().get_editor(object)

    def get_data_column(self, object):
        """ Gets the entire contents of the specified object column.
        """
        return getattr(object, self.name)

    def _is_selected(self, object):
        """ Returns whether a specified object row is selected.
        """
        if (
            hasattr(object, "model_selection")
            and object.model_selection is not None
        ):
            return True
        return False


class ListColumn(TableColumn):
    """ A column for editing lists.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    # Label to use for this column
    label = Property()

    #: Index of the list element associated with this column
    index = Int()

    # Is this column editable? This value overrides the base class default.
    editable = False

    # -------------------------------------------------------------------------
    #  Trait view definitions:
    # -------------------------------------------------------------------------

    traits_view = View(
        [
            ["index", "label", "type", "|[Column Information]"],
            ["text_color@", "cell_color@", "|[UI Colors]"],
        ]
    )

    def _get_label(self):
        """ Gets the label of the column.
        """
        if self._label is not None:
            return self._label
        return "Column %d" % (self.index + 1)

    def _set_label(self, label):
        old, self._label = self._label, label
        if old != label:
            self.trait_property_changed("label", old, label)

    def get_value(self, object):
        """ Gets the value of the column for a specified object.
        """
        return str(object[self.index])

    def set_value(self, object, value):
        """ Sets the value of the column for a specified object.
        """
        object[self.index] = value

    def get_editor(self, object):
        """ Gets the editor for the column of a specified object.
        """
        return None

    def key(self, object):
        """ Returns the value to use for sorting.
        """
        return object[self.index]



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# instance_editor_selection.py -- Example of an instance editor with instance
# selection

from traits.api import HasStrictTraits, Instance, Int, List, Regex, Str
from traitsui.api import InstanceEditor, Item, View


class Person(HasStrictTraits):
    name = Str()
    age = Int()
    phone = Regex(value='000-0000', regex=r'\d\d\d[-]\d\d\d\d')

    traits_view = View('name', 'age', 'phone')


people = [
    Person(name='Dave', age=39, phone='555-1212'),
    Person(name='Mike', age=28, phone='555-3526'),
    Person(name='Joe', age=34, phone='555-6943'),
    Person(name='Tom', age=22, phone='555-7586'),
    Person(name='Dick', age=63, phone='555-3895'),
    Person(name='Harry', age=46, phone='555-3285'),
    Person(name='Sally', age=43, phone='555-8797'),
    Person(name='Fields', age=31, phone='555-3547'),
]


class Team(HasStrictTraits):

    name = Str()
    captain = Instance(Person)
    roster = List(Person)

    traits_view = View(
        Item('name'),
        Item('_'),
        Item(
            'captain',
            label='Team Captain',
            editor=InstanceEditor(name='roster', editable=True),
            style='custom',
        ),
        buttons=['OK'],
    )


if __name__ == '__main__':
    team = Team(name='Vultures', captain=people[0], roster=people)
    team.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import warnings

import wx

from traitsui.testing.tester._ui_tester_registry._compat import (
    check_key_compat
)
from traitsui.testing.tester.exceptions import Disabled


def _create_event(event_type, control):
    """ Creates a wxEvent of a given type

    Parameters
    ----------
    event_type : wxEventType
        The type of the event to be created
    control :
        The wx control the event is occurring on.

    Returns
    -------
    wxEvent
        The created event, of the given type, with the given control set as
        the event object.
    """
    event = wx.CommandEvent(
        event_type, control.GetId()
    )
    event.SetEventObject(control)
    return event


def mouse_click(func):
    """ Decorator function for mouse clicks. Decorated functions will return
    if they are not enabled. Additionally, this handles the delay for the
    click.

    Parameters
    ----------
    func : callable(*, control, delay, **kwargs)
        The mouse click function to be decorated.

    Returns
    -------
    callable
        The decorated function.
    """
    def mouse_click_handler(*, control, delay, **kwargs):
        """ Defines the decorated function.

        Paramters
        ---------
        control : wxControl
            The wx Object to be clicked.
        delay : int
            Time delay (in ms) in which click will be performed.
        """
        if (not control) or (not control.IsEnabled()):
            warnings.warn(
                "Attempted to click on a non-existant or non-enabled control. "
                "Nothing was performed."
            )
            return
        wx.MilliSleep(delay)
        func(control=control, delay=delay, **kwargs)
    return mouse_click_handler


@mouse_click
def mouse_click_button(control, delay):
    """ Performs a mouce click on a wx button.

    Parameters
    ----------
    control : wxButton
        The wx Object to be clicked.
    delay: int
        Time delay (in ms) in which click will be performed.
    """
    click_event = _create_event(wx.wxEVT_COMMAND_BUTTON_CLICKED, control)
    control.ProcessWindowEvent(click_event)


@mouse_click
def mouse_click_checkbox(control, delay):
    """ Performs a mouce click on a wx check box.

    Parameters
    ----------
    control : wxCheckBox
        The wx Object to be clicked.
    delay: int
        Time delay (in ms) in which click will be performed.
    """
    click_event = _create_event(wx.wxEVT_COMMAND_CHECKBOX_CLICKED, control)
    control.SetValue(not control.GetValue())
    control.ProcessWindowEvent(click_event)


@mouse_click
def mouse_click_combobox_or_choice(control, index, delay):
    """ Performs a mouce click on either a wx combo box or a wx choice on the
    entry at the given index.

    Parameters
    ----------
    control : wxComboBox or wxChoice
        The wx Object to be clicked.
    index : int
        The index of the item in the combobox/choice to be clicked
    delay: int
        Time delay (in ms) in which click will be performed.

    Raises
    ------
    TypeError
        If the control is not a wxComboBox or wxChoice.
    """
    if isinstance(control, wx.ComboBox):
        click_event = _create_event(
            wx.wxEVT_COMMAND_COMBOBOX_SELECTED, control
        )
    elif isinstance(control, wx.Choice):
        click_event = _create_event(wx.wxEVT_COMMAND_CHOICE_SELECTED, control)
    else:
        raise TypeError("Only supported controls are wxComboBox or wxChoice")
    click_event.SetString(control.GetString(index))
    control.SetSelection(index)
    control.ProcessWindowEvent(click_event)


@mouse_click
def mouse_click_listbox(control, index, delay):
    """Performs a mouce click on a wx list box on the entry at
    the given index.

    Parameters
    ----------
    control : wxListBox
        The wx Object to be clicked.
    index : int
        The index of the item in the list box to be clicked
    delay: int
        Time delay (in ms) in which click will be performed.
    """
    click_event = _create_event(wx.wxEVT_COMMAND_LISTBOX_SELECTED, control)
    control.SetSelection(index)
    control.ProcessWindowEvent(click_event)


@mouse_click
def mouse_click_radiobutton(control, delay):
    """ Performs a mouce click on a wx radio button.

    Parameters
    ----------
    control : wxRadioButton
        The wx Object to be clicked.
    delay: int
        Time delay (in ms) in which click will be performed.
    """
    click_event = _create_event(wx.wxEVT_COMMAND_RADIOBUTTON_SELECTED, control)
    control.SetValue(not control.GetValue())
    control.ProcessWindowEvent(click_event)


@mouse_click
def mouse_click_object(control, delay):
    """ Performs a mouce click on a wxTextCtrl.

    Parameters
    ----------
    control : wxObject
        The wx Object to be clicked.
    delay: int
        Time delay (in ms) in which click will be performed.
    """
    if not control.HasFocus():
        control.SetFocus()
    click_event = _create_event(wx.wxEVT_COMMAND_LEFT_CLICK, control)
    control.ProcessWindowEvent(click_event)


def mouse_click_notebook_tab_index(control, index, delay):
    """ Performs a mouseclick on a Noteboook List Editor on the tab specified
    by index.

    Parameters
    ----------
    control : wx.Window
        The control of the DockWindow
    index : int
        The index of the child object in the Panel to be clicked
    delay : int
        Time delay (in ms) in which click will be performed.
    """
    controls_list = control.GetSizer().GetContents().get_controls()
    wx.MilliSleep(delay)

    # find the boundaries of the tab to be clicked
    bx, by, bdx, bdy = controls_list[index].drag_bounds
    # find the center
    tab_center = wx.Point(bx + bdx//2, by + bdy//2)

    click_down_event = wx.MouseEvent(
        wx.wxEVT_LEFT_DOWN
    )
    click_down_event.SetPosition(tab_center)
    click_up_event = wx.MouseEvent(
        wx.wxEVT_LEFT_UP
    )
    click_up_event.SetPosition(tab_center)
    control.ProcessEvent(click_down_event)
    control.ProcessEvent(click_up_event)


def mouse_click_checkbox_child_in_panel(control, index, delay):
    """ Performs a mouce click on a child of a Wx Panel.

    Parameters
    ----------
    control : wx.Panel
        The Panel containing child objects, one of which will be clicked.
    index : int
        The index of the child object in the Panel to be clicked
    delay : int
        Time delay (in ms) in which click will be performed.

    Raises
    ------
    IndexError
        If the index is out of range.
    """
    children_list = control.GetSizer().GetChildren()
    if not 0 <= index <= len(children_list) - 1:
        raise IndexError(index)
    obj = children_list[index].GetWindow()
    mouse_click_checkbox(control=obj, delay=delay)


def mouse_click_radiobutton_child_in_panel(control, index, delay):
    """ Performs a mouce click on a child of a Wx Panel.

    Parameters
    ----------
    control : wx.Panel
        The Panel containing child objects, one of which will be clicked.
    index : int
        The index of the child object in the Panel to be clicked
    delay : int
        Time delay (in ms) in which click will be performed.

    Raises
    ------
    IndexError
        If the index is out of range.
    """
    children_list = control.GetSizer().GetChildren()
    if not 0 <= index <= len(children_list) - 1:
        raise IndexError(index)
    obj = children_list[index].GetWindow()
    mouse_click_radiobutton(control=obj, delay=delay)


def key_click_text_entry(
    control, interaction, delay,
    get_selection=lambda control: control.GetSelection()
):
    """ Performs simulated typing of a key on the given wxObject
    after a delay.

    Parameters
    ----------
    control : wxTextEntry
        The wx Object to be acted on.
    interaction : instance of command.KeyClick
        The interaction object holding the key input
        to be simulated being typed
    delay : int
        Time delay (in ms) in which the key click will be performed.
    get_selection: callable(wx.TextEntry) -> tuple(int, int)
        Callable that takes an instance of wx.TextEntry and return the
        current selection span. Default is to call `GetSelection` method.
        Useful for when the TextEntry.GetSelection is overridden by a subclass
        that does not conform to the common API.
    """
    if not (control.IsEnabled() and control.IsEditable()):
        raise Disabled("{!r} is disabled.".format(control))
    if not control.HasFocus():
        control.SetFocus()
        control.SetInsertionPointEnd()
    if interaction.key == "Enter":
        wx.MilliSleep(delay)
        event = wx.CommandEvent(wx.EVT_TEXT_ENTER.typeId, control.GetId())
        control.ProcessEvent(event)
    elif interaction.key == "End":
        wx.MilliSleep(delay)
        control.SetInsertionPointEnd()
    elif interaction.key == "Backspace":
        wx.MilliSleep(delay)
        start, end = get_selection(control)
        if end > start:
            control.Remove(start, end)
        else:
            pos = control.GetInsertionPoint()
            control.Remove(max(0, pos - 1), pos)
    else:
        check_key_compat(interaction.key)
        wx.MilliSleep(delay)
        control.WriteText(interaction.key)


def key_click_combobox(control, interaction, delay):
    """ Performs simulated typing of a key on the given wxComboBox
    after a delay.

    Parameters
    ----------
    control : wxComboBox
        The wx Object to be acted on.
    interaction : instance of command.KeyClick
        The interaction object holding the key input
        to be simulated being typed
    delay : int
        Time delay (in ms) in which the key click will be performed.
    """
    key_click_text_entry(
        control, interaction, delay,
        get_selection=lambda control: control.GetTextSelection(),
    )


def key_sequence_text_ctrl(control, interaction, delay):
    """ Performs simulated typing of a sequence of keys on the given wxObject
    after a delay.

    Parameters
    ----------
    control : wxTextCtrl
        The wx Object to be acted on.
    interaction : instance of command.KeySequence
        The interaction object holding the sequence of key inputs
        to be simulated being typed
    delay : int
        Time delay (in ms) in which each key click in the sequence will be
        performed.

    Raises
    ------
    Disabled
        If the control is either not enabled or not editable.
    """
    # fail early
    for char in interaction.sequence:
        check_key_compat(char)

    if not (control.IsEnabled() and control.IsEditable()):
        raise Disabled("{!r} is disabled.".format(control))
    if not control.HasFocus():
        control.SetFocus()
        control.SetInsertionPointEnd()
    for char in interaction.sequence:
        wx.MilliSleep(delay)
        control.WriteText(char)
        wx.SafeYield()


def key_click_slider(control, interaction, delay):
    """ Performs simulated typing of a key on the given wxSlider
    after a delay. Only allowed keys are:
    "Left", "Right", "Up", "Down", "Page Up", "Page Down"
    Also, note that up related keys correspond to an increment on the slider,
    and down a decrement.

    Parameters
    ----------
    control : wxSlider
        The wx Object to be acted on.
    interaction : instance of command.KeyClick
        The interaction object holding the key input
        to be simulated being typed
    delay : int
        Time delay (in ms) in which the key click will be performed.

    Raises
    ------
    ValueError
        If the interaction.key is not one of the valid keys.
    """
    valid_keys = {"Left", "Right", "Up", "Down", "Page Up", "Page Down"}
    if interaction.key not in valid_keys:
        raise ValueError(
            "Unexpected Key. Supported keys are: {}".format(sorted(valid_keys))
        )
    if not control.HasFocus():
        control.SetFocus()
    value = control.GetValue()
    if interaction.key in {"Up", "Right"}:
        position = min(control.GetMax(), value + control.GetLineSize())
    elif interaction.key == "Page Up":
        position = min(control.GetMax(), value + control.GetPageSize())
    elif interaction.key == "Page Down":
        position = max(control.GetMin(), value - control.GetPageSize())
    elif interaction.key in {"Down", "Left"}:
        position = max(control.GetMin(), value - control.GetLineSize())
    else:
        raise ValueError(
            "Unexpected Key. Supported keys are: {}".format(sorted(valid_keys))
        )
    wx.MilliSleep(delay)
    control.SetValue(position)
    event = wx.ScrollEvent(
        wx.wxEVT_SCROLL_CHANGED, control.GetId(), position
    )
    wx.PostEvent(control, event)


def readonly_textbox_displayed_text(control):
    ''' Extracts the displayed text in a wx textbox (either a wx.TextCtrl or
    wx.StaticText).

    Parameters
    ----------
    control : wx.TextCtrl or wx.StaticText
        the textbox object from which the text of interest is displayed

    Raises
    ------
    TypeError
        If the control is not either a wx.TextCtrl or wx.StaticText.
    '''
    if isinstance(control, wx.TextCtrl):
        return control.GetValue()
    elif isinstance(control, wx.StaticText):
        return control.GetLabel()
    raise TypeError("readonly_textbox_displayed_text expected a control"
                    " of either wx.TextCtrl, or wx.StaticText."
                    " {} was found".format(control))



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import logging

from traits.api import (
    Bool,
    Dict,
    Enum,
    Instance,
    List,
    Property,
    Str,
    Tuple,
    Either,
)

from traitsui.basic_editor_factory import BasicEditorFactory
from traitsui.editors.tabular_editor import TabularEditor
from traitsui.item import Item
from traitsui.tabular_adapter import TabularAdapter
from traitsui.toolkit import toolkit_object
from traitsui.toolkit_traits import Font
from traitsui.ui_editor import UIEditor
from traitsui.view import View



logger = logging.getLogger(__name__)


class DataFrameAdapter(TabularAdapter):
    """ Generic tabular adapter for data frames
    """

    #: The text to use for a generic entry.
    text = Property()

    #: The alignment for each cell
    alignment = Property(Enum("left", "center", "right"))

    #: The text to use for a row index.
    index_text = Property()

    #: The alignment to use for a row index.
    index_alignment = Property()

    #: The font to use for each column
    font = Property()

    #: The format to use for each column
    format = Property()

    #: The format for each element, or a mapping column ID to format.
    _formats = Either(Str, Dict, default="%s")

    #: The font for each element, or a mapping column ID to font.
    _fonts = Either(Font, Dict, default="Courier 10")

    def _get_index_alignment(self):
        import numpy as np

        index = getattr(self.object, self.name).index
        if np.issubdtype(index.dtype, np.number):
            return "right"
        else:
            return "left"

    def _get_alignment(self):
        import numpy as np

        column = self.item[self.column_id]
        if np.issubdtype(column.dtype, np.number):
            return "right"
        else:
            return "left"

    def _get_font(self):
        if isinstance(self._fonts, toolkit_object("font_trait:TraitsFont")):
            return self._fonts
        else:
            return self._fonts.get(self.column_id, "Courier 10")

    def _get_format(self):
        if isinstance(self._formats, str):
            return self._formats
        else:
            return self._formats.get(self.column_id, "%s")

    def _get_content(self):
        return self.item[self.column_id].iloc[0]

    def _get_text(self):
        format = self.get_format(self.object, self.name, self.row, self.column)
        return format % self.get_content(
            self.object, self.name, self.row, self.column
        )

    def _set_text(self, value):
        df = getattr(self.object, self.name)
        dtype = df.iloc[:, self.column].dtype
        try:
            value = dtype.type(value)
            df.iloc[self.row, self.column] = value
        except Exception:
            logger.debug(
                "User entered invalid value %r for column %r row %r",
                value,
                self.column,
                self.row,
                exc_info=True
            )

    def _get_index_text(self):
        return str(self.item.index[0])

    def _set_index_text(self, value):
        index = getattr(self.object, self.name).index
        dtype = index.dtype
        try:
            value = dtype.type(value)
            index.values[self.row] = value
        except Exception:
            logger.debug(
                "User entered invalid value %r for index",
                value,
                exc_info=True
            )

    # ---- Adapter methods that are not sensitive to item type ----------------

    def get_item(self, object, trait, row):
        """ Override the base implementation to work with DataFrames

        This returns a dataframe with one row, rather than a series, since
        using a dataframe preserves dtypes.

        """
        return getattr(object, trait).iloc[row : row + 1]

    def delete(self, object, trait, row):
        """ Override the base implementation to work with DataFrames

        Unavoidably does a copy of the data, setting the trait with the new
        value.
        """
        import pandas as pd

        df = getattr(object, trait)
        if 0 < row < len(df) - 1:
            new_df = pd.concat([df.iloc[:row, :], df.iloc[row + 1 :, :]])
        elif row == 0:
            new_df = df.iloc[row + 1 :, :]
        else:
            new_df = df.iloc[:row, :]
        setattr(object, trait, new_df)

    def insert(self, object, trait, row, value):
        """ Override the base implementation to work with DataFrames

        Unavoidably does a copy of the data, setting the trait with the new
        value.
        """
        import pandas as pd

        df = getattr(object, trait)
        if 0 < row < len(df) - 1:
            new_df = pd.concat([df.iloc[:row, :], value, df.iloc[row:, :]])
        elif row == 0:
            new_df = pd.concat([value, df])
        else:
            new_df = pd.concat([df, value])
        setattr(object, trait, new_df)


class _DataFrameEditor(UIEditor):
    """ TraitsUI-based editor implementation for data frames """

    #: Indicate that the editor is scrollable/resizable:
    scrollable = True

    #: Should column titles be displayed:
    show_titles = Bool(True)

    #: The tabular adapter being used for the editor view:
    adapter = Instance(DataFrameAdapter)

    # -- Private Methods ------------------------------------------------------

    def _target_name(self, name):
        if name:
            return "object.object." + name
        else:
            return ""

    def _data_frame_view(self):
        """ Return the view used by the editor.
        """

        return View(
            Item(
                self._target_name(self.name),
                id="tabular_editor",
                show_label=False,
                editor=TabularEditor(
                    show_titles=self.factory.show_titles,
                    editable=self.factory.editable,
                    adapter=self.adapter,
                    selected=self._target_name(self.factory.selected),
                    selected_row=self._target_name(self.factory.selected_row),
                    selectable=self.factory.selectable,
                    multi_select=self.factory.multi_select,
                    activated=self._target_name(self.factory.activated),
                    activated_row=self._target_name(
                        self.factory.activated_row
                    ),  # noqa
                    clicked=self._target_name(self.factory.clicked),
                    dclicked=self._target_name(self.factory.dclicked),
                    scroll_to_row=self._target_name(
                        self.factory.scroll_to_row
                    ),  # noqa
                    scroll_to_position_hint=(
                        self.factory.scroll_to_position_hint
                    ),
                    scroll_to_column=self._target_name(
                        self.factory.scroll_to_column
                    ),  # noqa
                    right_clicked=self._target_name(
                        self.factory.right_clicked
                    ),  # noqa
                    right_dclicked=self._target_name(
                        self.factory.right_dclicked
                    ),  # noqa
                    column_clicked=self._target_name(
                        self.factory.column_clicked
                    ),  # noqa
                    column_right_clicked=self._target_name(
                        self.factory.column_right_clicked
                    ),  # noqa
                    operations=self.factory.operations,
                    update=self._target_name(self.factory.update),
                    refresh=self._target_name(self.factory.refresh),
                ),
            ),
            id="data_frame_editor",
            resizable=True,
        )

    def init_ui(self, parent):
        """ Creates the Traits UI for displaying the array.
        """
        factory = self.factory
        if factory.columns != []:
            columns = []
            for column in factory.columns:
                if isinstance(column, str):
                    title = column
                    column_id = column
                else:
                    title, column_id = column
                if column_id not in self.value.columns:
                    continue
                columns.append((title, column_id))
        else:
            columns = [
                (column_id, column_id) for column_id in self.value.columns
            ]

        if factory.show_index:
            index_name = self.value.index.name
            if index_name is None:
                index_name = ""
            columns.insert(0, (index_name, "index"))

        if factory.adapter is not None:
            self.adapter = factory.adapter
            self.adapter._formats = factory.formats
            self.adapter._fonts = factory.fonts
            if not self.adapter.columns:
                self.adapter.columns = columns
        else:
            self.adapter = DataFrameAdapter(
                columns=columns, _formats=factory.formats, _fonts=factory.fonts
            )

        return self.edit_traits(
            view="_data_frame_view", parent=parent, kind="subpanel"
        )


class DataFrameEditor(BasicEditorFactory):
    """ Editor factory for basic data frame editor """

    #: The editor implementation class.
    klass = Property()

    #: Should an index column be displayed.
    show_index = Bool(True)

    #: Should column headers be displayed.
    show_titles = Bool(True)

    #: Optional list of either column ID or pairs of (column title, column ID).
    columns = List(Either(Str, Tuple(Str, Str)))

    #: The format for each element, or a mapping column ID to format.
    formats = Either(Str, Dict, default="%s")

    #: The font for each element, or a mapping column ID to font.
    fonts = Either(Font, Dict, default="Courier 10")

    #: The optional extended name of the trait to synchronize the selection
    #: values with:
    selected = Str()

    #: The optional extended name of the trait to synchronize the selection rows
    #: with:
    selected_row = Str()

    #: Whether or not to allow selection.
    selectable = Bool(True)

    #: Whether or not to allow for multiple selections
    multi_select = Bool(False)

    #: The optional extended name of the trait to synchronize the activated
    #: value with:
    activated = Str()

    #: The optional extended name of the trait to synchronize the activated
    #: value's row with:
    activated_row = Str()

    #: The optional extended name of the trait to synchronize left click data
    #: with. The data is a TabularEditorEvent:
    clicked = Str()

    #: The optional extended name of the trait to synchronize left double click
    #: data with. The data is a TabularEditorEvent:
    dclicked = Str()

    #: The optional extended name of the Event trait that should be used to
    #: trigger a scroll-to command. The data is an integer giving the row.
    scroll_to_row = Str()

    #: Deprecated: Controls behavior of scroll to row and scroll to column
    scroll_to_row_hint = Property(Str, observe="scroll_to_position_hint")

    #: (replacement of scroll_to_row_hint, but more clearly named)
    #: Controls behavior of scroll to row and scroll to column
    scroll_to_position_hint = Enum("visible", "center", "top", "bottom")

    #: The optional extended name of the Event trait that should be used to
    #: trigger a scroll-to command. The data is an integer giving the column.
    scroll_to_column = Str()

    #: The optional extended name of the trait to synchronize right click data
    #: with. The data is a TabularEditorEvent:
    right_clicked = Str()

    #: The optional extended name of the trait to synchronize right double
    #: clicked data with. The data is a TabularEditorEvent:
    right_dclicked = Str()

    #: The optional extended name of the trait to synchronize column
    #: clicked data with. The data is a TabularEditorEvent:
    column_clicked = Str()

    #: The optional extended name of the trait to synchronize column
    #: right clicked data with. The data is a TabularEditorEvent:
    column_right_clicked = Str()

    #: Whether or not the entries can be edited.
    editable = Bool(False)

    #: What type of operations are allowed on the list:
    operations = List(
        Enum("delete", "insert", "append", "edit", "move"),
        ["delete", "insert", "append", "edit", "move"],
    )

    #: The optional extended name of the trait used to indicate that a complete
    #: table update is needed:
    update = Str()

    #: The optional extended name of the trait used to indicate that the table
    #: just needs to be repainted.
    refresh = Str()

    #: Set to override the default dataframe adapter
    adapter = Instance(DataFrameAdapter)

    def _get_klass(self):
        """ The class used to construct editor objects.
        """
        return toolkit_object("data_frame_editor:_DataFrameEditor")
    
    def _get_scroll_to_row_hint(self):
        warnings.warn(
            "Use of scroll_to_row_hint trait is deprecated. "
            "Use scroll_to_position_hint instead.",
            DeprecationWarning,
        )
        return self.scroll_to_position_hint
    
    def _set_scroll_to_row_hint(self, hint):
        warnings.warn(
            "Use of scroll_to_row_hint trait is deprecated. "
            "Use scroll_to_position_hint instead.",
            DeprecationWarning,
        )
        self.scroll_to_position_hint = hint



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Tests for traitsui.testing._exception_handling """

import unittest

from pyface.api import GUI

from traitsui.tests._tools import (
    requires_toolkit,
    ToolkitName,
)
from traitsui.testing._exception_handling import reraise_exceptions


class TestExceptionHandling(unittest.TestCase):

    @requires_toolkit([ToolkitName.qt, ToolkitName.wx])
    def test_error_from_gui_captured_and_raise(self):

        def raise_error_1():
            raise ZeroDivisionError()

        def raise_error_2():
            raise IndexError()

        # without the context manager:
        #   - with Qt5, the test run will be aborted prematurely.
        #   - with Qt4, the traceback is printed and the test passes.
        #   - with Wx, the traceback is printed and the test passes.
        # With the context manager, the exception is always reraised.
        gui = GUI()
        with self.assertRaises(RuntimeError) as exception_context, \
                self.assertLogs("traitsui") as watcher:
            with reraise_exceptions():
                gui.invoke_later(raise_error_1)
                gui.invoke_later(raise_error_2)
                gui.process_events()

        error_msg = str(exception_context.exception)
        self.assertIn("ZeroDivisionError", error_msg)
        self.assertIn("IndexError", error_msg)
        log_content1, log_content2 = watcher.output
        self.assertIn("ZeroDivisionError", log_content1)
        self.assertIn("IndexError", log_content2)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

Defining column-specific context menu in a Tabular Editor. Shows how the
example for the Table Editor (`Table_Editor_with_context_menu_demo.py`) can be
adapted to work with a Tabular Editor.

The demo is a simple baseball scoring system, which lists each player and
their current batting statistics. After a given player has an at bat, you
right-click on the table cell corresponding to the player and the result of
the at-bat (e.g. 'S' = single) and select the 'Add' menu option to register
that the player hit a single and update the player's overall statistics.

This demo also illustrates the use of Property traits, and how using 'event'
meta-data can simplify event handling by collapsing an event that can
occur on a number of traits into a category of event, which can be handled by
a single event handler defined for the category (in this case, the category
is 'affects_average').
"""
# Issue related to the demo warning: enthought/traitsui#960


from random import randint
from traits.api import HasStrictTraits, Str, Int, Float, List, Property
from traitsui.api import (
    Action, Item, Menu, TabularAdapter, TabularEditor, View
)


# Define a custom tabular adapter for handling items which affect the player's
# batting average:
class PlayerAdapter(TabularAdapter):

    # Overwrite default values
    alignment = 'center'
    width = 0.09

    def get_menu(self, object, trait, row, column):
        column_name = self.column_map[column]
        if column_name not in ['name', 'average']:
            menu = Menu(
                Action(
                    name='Add',
                    action='editor.adapter.add(item, column)'
                ),
                Action(
                    name='Sub',
                    action='editor.adapter.sub(item, column)'
                )
            )
            return menu
        else:
            return super().get_menu(object, trait, row, column)

    def get_format(self, object, trait, row, column):
        column_name = self.column_map[column]
        if column_name == 'average':
            return '%0.3f'
        else:
            return super().get_format(object, trait, row, column)

    def add(self, object, column):
        """ Increment the affected player statistic.
        """
        column_name = self.column_map[column]
        setattr(object, column_name, getattr(object, column_name) + 1)

    def sub(self, object, column):
        """ Decrement the affected player statistic.
        """
        column_name = self.column_map[column]
        setattr(object, column_name, getattr(object, column_name) - 1)


# The 'players' trait table editor:
columns = [
    ('Player Name', 'name'),
    ('AB', 'at_bats'),
    ('SO', 'strike_outs'),
    ('S', 'singles'),
    ('D', 'doubles'),
    ('T', 'triples'),
    ('HR', 'home_runs'),
    ('W', 'walks'),
    ('Ave', 'average'),
]

player_editor = TabularEditor(
    editable=True,
    auto_resize=True,
    auto_resize_rows=True,
    stretch_last_section=False,
    auto_update=True,
    adapter=PlayerAdapter(columns=columns),
)


# 'Player' class:
class Player(HasStrictTraits):

    # Trait definitions:
    name = Str()
    at_bats = Int()
    strike_outs = Int(event='affects_average')
    singles = Int(event='affects_average')
    doubles = Int(event='affects_average')
    triples = Int(event='affects_average')
    home_runs = Int(event='affects_average')
    walks = Int()
    average = Property(Float)

    def _get_average(self):
        """ Computes the player's batting average from the current statistics.
        """
        if self.at_bats == 0:
            return 0.0

        return float(self.singles + self.doubles +
                     self.triples + self.home_runs) / self.at_bats

    def _affects_average_changed(self):
        """ Handles an event that affects the player's batting average.
        """
        self.at_bats += 1


class Team(HasStrictTraits):

    # Trait definitions:
    players = List(Player)

    # Trait view definitions:
    traits_view = View(
        Item('players', show_label=False, editor=player_editor),
        title='Baseball Scoring Demo',
        width=0.5,
        height=0.5,
        resizable=True
    )


def random_player(name):
    """ Generates and returns a random player.
    """
    p = Player(
        name=name,
        strike_outs=randint(0, 50),
        singles=randint(0, 50),
        doubles=randint(0, 20),
        triples=randint(0, 5),
        home_runs=randint(0, 30),
        walks=randint(0, 50)
    )
    return p.trait_set(
        at_bats=(
            p.strike_outs + p.singles + p.doubles + p.triples + p.home_runs +
            randint(100, 200)
        )
    )


# Create the demo:
demo = Team(
    players=[
        random_player(name) for name in [
            'Dave', 'Mike', 'Joe', 'Tom', 'Dick', 'Harry', 'Dirk', 'Fields',
            'Stretch'
        ]
    ]
)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

Implementation of a CompoundEditor demo plugin for Traits UI demo program.

This demo shows each of the four styles of the CompoundEditor.

Please refer to the `CompoundEditor API docs`_ for further information.

.. _CompoundEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.compound_editor.html#traitsui.editors.compound_editor.CompoundEditor
"""
# Issue related to the demo warning: enthought/traitsui#945


from traits.api import Either, Enum, HasTraits, Range

from traitsui.api import Item, Group, View


# Define the demo class:
class CompoundEditorDemo(HasTraits):
    """ Defines the main CompoundEditor demo class.
    """

    # Define a compund trait to view:
    # Note: In Traits 6.1 a new Trait type `Union` has been added, which is
    # recommended over `Either`:
    # compound_trait = Union(
    #     Range(1, 6), Enum('a', 'b', 'c', 'd', 'e', 'f'), default_value=1
    # )
    compound_trait = Either(
        Range(1, 6), Enum('a', 'b', 'c', 'd', 'e', 'f'), default=1
    )

    # Display specification (one Item per editor style):
    comp_group = Group(
        Item('compound_trait', style='simple', label='Simple'),
        Item('_'),
        Item('compound_trait', style='custom', label='Custom'),
        Item('_'),
        Item('compound_trait', style='text', label='Text'),
        Item('_'),
        Item('compound_trait', style='readonly', label='ReadOnly')
    )

    # Demo view:
    traits_view = View(
        comp_group,
        title='CompoundEditor',
        buttons=['OK'],
        resizable=True
    )


# Create the demo:
demo = CompoundEditorDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The organization in this package should more or less mirror that of
the ``traitsui.wx`` package and contains the corresponding logic for testing.

Note that once new implementations are added, they may need to be exposed
via the ``default_registry`` module (one level up from this package).
"""



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the various editors for a drag-and-drop editor,
    for the wxPython user interface toolkit. A drag-and-drop editor represents
    its value as a simple image which, depending upon the editor style, can be
    a drag source only, a drop target only, or both a drag source and a drop
    target.
"""


import wx
import numpy

from pickle import load

from traits.api import Bool

from pyface.wx.drag_and_drop import (
    PythonDropSource,
    PythonDropTarget,
    clipboard,
)


try:
    from apptools.io import File
except ImportError:
    File = None

try:
    from apptools.naming.api import Binding
except ImportError:
    Binding = None

from pyface.image_resource import ImageResource

from .editor import Editor


# The image to use when the editor accepts files:
file_image = ImageResource("file").create_image()

# The image to use when the editor accepts objects:
object_image = ImageResource("object").create_image()

# The image to use when the editor is disabled:
inactive_image = ImageResource("inactive").create_image()

# String types:
string_type = (str, str)

# -------------------------------------------------------------------------
#  'SimpleEditor' class:
# -------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simply style of editor for a drag-and-drop editor, which is both a drag
        source and a drop target.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Is the editor a drop target?
    drop_target = Bool(True)

    #: Is the editor a drag source?
    drag_source = Bool(True)

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        # Determine the drag/drop type:
        value = self.value
        self._is_list = isinstance(value, list)
        self._is_file = isinstance(value, string_type) or (
            self._is_list
            and (len(value) > 0)
            and isinstance(value[0], string_type)
        )

        # Get the right image to use:
        image = self.factory.image
        if image is not None:
            image = image.create_image()
            disabled_image = self.factory.disabled_image
            if disabled_image is not None:
                disabled_image = disabled_image.create_image()
        else:
            disabled_image = inactive_image
            image = object_image
            if self._is_file:
                image = file_image

        self._image = image.ConvertToBitmap()
        if disabled_image is not None:
            self._disabled_image = disabled_image.ConvertToBitmap()
        else:
            data = numpy.reshape(
                numpy.fromstring(image.GetData(), numpy.uint8), (-1, 3)
            ) * numpy.array([[0.297, 0.589, 0.114]])
            g = data[:, 0] + data[:, 1] + data[:, 2]
            data[:, 0] = data[:, 1] = data[:, 2] = g
            image.SetData(numpy.ravel(data.astype(numpy.uint8)).tostring())
            image.SetMaskColour(0, 0, 0)
            self._disabled_image = image.ConvertToBitmap()

        # Create the control and set up the event handlers:
        self.control = control = wx.Window(
            parent, -1, size=wx.Size(image.GetWidth(), image.GetHeight())
        )
        self.set_tooltip()

        if self.drop_target:
            control.SetDropTarget(PythonDropTarget(self))

        control.Bind(wx.EVT_LEFT_DOWN, self._left_down)
        control.Bind(wx.EVT_LEFT_UP, self._left_up)
        control.Bind(wx.EVT_MOTION, self._mouse_move)
        control.Bind(wx.EVT_PAINT, self._on_paint)

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        control = self.control
        control.Unbind(wx.EVT_LEFT_DOWN)
        control.Unbind(wx.EVT_LEFT_UP)
        control.Unbind(wx.EVT_MOTION)
        control.Unbind(wx.EVT_PAINT)

        super().dispose()

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        return

    # -- Private Methods ------------------------------------------------------

    def _get_drag_data(self, data):
        """ Returns the processed version of a drag request's data.
        """
        if isinstance(data, list):

            if Binding is not None and isinstance(data[0], Binding):
                data = [item.obj for item in data]

            if File is not None and isinstance(data[0], File):
                data = [item.absolute_path for item in data]
                if not self._is_file:
                    result = []
                    for file in data:
                        item = self._unpickle(file)
                        if item is not None:
                            result.append(item)
                    data = result

        else:
            if Binding is not None and isinstance(data, Binding):
                data = data.obj

            if File is not None and isinstance(data, File):
                data = data.absolute_path
                if not self._is_file:
                    object = self._unpickle(data)
                    if object is not None:
                        data = object

        return data

    def _unpickle(self, file_name):
        """ Returns the unpickled version of a specified file (if possible).
        """
        with open(file_name, "rb") as fh:
            try:
                object = load(fh)
            except Exception:
                object = None

        return object

    # -- wxPython Event Handlers ----------------------------------------------

    def _on_paint(self, event):
        """ Called when the control needs repainting.
        """
        image = self._image
        control = self.control
        if not control.IsEnabled():
            image = self._disabled_image

        wdx, wdy = control.GetClientSize()
        wx.PaintDC(control).DrawBitmap(
            image,
            (wdx - image.GetWidth()) // 2,
            (wdy - image.GetHeight()) // 2,
            True,
        )

    def _left_down(self, event):
        """ Handles the left mouse button being pressed.
        """
        if self.control.IsEnabled() and self.drag_source:
            self._x, self._y = event.GetX(), event.GetY()
            self.control.CaptureMouse()

        event.Skip()

    def _left_up(self, event):
        """ Handles the left mouse button being released.
        """
        if self._x is not None:
            self._x = None
            self.control.ReleaseMouse()

        event.Skip()

    def _mouse_move(self, event):
        """ Handles the mouse being moved.
        """
        if self._x is not None:
            if (
                abs(self._x - event.GetX()) + abs(self._y - event.GetY())
            ) >= 3:
                self.control.ReleaseMouse()
                self._x = None
                if self._is_file:
                    FileDropSource(self.control, self.value)
                else:
                    PythonDropSource(self.control, self.value)

        event.Skip()

    # ----- Drag and drop event handlers: -------------------------------------

    def wx_dropped_on(self, x, y, data, drag_result):
        """ Handles a Python object being dropped on the tree.
        """
        try:
            self.value = self._get_drag_data(data)
            return drag_result
        except:
            return wx.DragNone

    def wx_drag_over(self, x, y, data, drag_result):
        """ Handles a Python object being dragged over the tree.
        """
        try:
            self.object.base_trait(self.name).validate(
                self.object, self.name, self._get_drag_data(data)
            )
            return drag_result
        except:
            return wx.DragNone


class CustomEditor(SimpleEditor):
    """ Custom style of drag-and-drop editor, which is not a drag source.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Is the editor a drag source? This value overrides the default.
    drag_source = False


class ReadonlyEditor(SimpleEditor):
    """ Read-only style of drag-and-drop editor, which is not a drop target.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Is the editor a drop target? This value overrides the default.
    drop_target = False


class FileDropSource(wx.DropSource):
    """ Represents a draggable file.
    """

    def __init__(self, source, files):
        """ Initializes the object.
        """
        self.handler = None
        self.allow_move = True

        # Put the data to be dragged on the clipboard:
        clipboard.data = files
        clipboard.source = source
        clipboard.drop_source = self

        data_object = wx.FileDataObject()
        if isinstance(files, string_type):
            files = [files]

        for file in files:
            data_object.AddFile(file)

        # Create the drop source and begin the drag and drop operation:
        super().__init__(source)
        self.SetData(data_object)
        self.result = self.DoDragDrop(True)

    def on_dropped(self, drag_result):
        """ Called when the data has been dropped. """
        return



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Implementation of a DateRangeEditor demo plugin for Traits UI demo program.

This demo shows a custom style DateRangeEditor. Note that this demo only works
with qt backend.
"""
# Issue related to the demo warning: enthought/traitsui#962

from traits.api import HasTraits, Date, Tuple
from traits.etsconfig.api import ETSConfig
from traitsui.api import View, Item, DateRangeEditor, Group


class DateRangeEditorDemo(HasTraits):
    """ Demo class to show DateRangeEditor. """
    date_range = Tuple(Date, Date)

    traits_view = View(
        Group(
            Item(
                'date_range',
                editor=DateRangeEditor(),
                style='custom',
                label='Date range'
            ),
            label='Date range'
        ),
        resizable=True
    )

    def _date_range_changed(self):
        print(self.date_range)


# -- Set Up The Demo ------------------------------------------------------


if __name__ == "__main__":
    if ETSConfig.toolkit == "qt4":
        # DateRangeEditor is currently only available for qt backend.
        demo = DateRangeEditorDemo()
        demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import HasTraits, Int
from traitsui.tests._tools import BaseTestMixin, requires_toolkit, ToolkitName


class TestNullToolkit(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @requires_toolkit([ToolkitName.null])
    def test_configure_traits_error(self):
        """ Verify that configure_traits fails with NotImplementedError. """

        class Test(HasTraits):
            x = Int()

        t = Test()

        with self.assertRaises(NotImplementedError):
            t.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import datetime
import unittest

from traits.api import Dict, HasTraits, Instance, List, Str
from traitsui.api import Item, StyledDateEditor, View
from traitsui.editors.date_editor import CellFormat
from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    requires_toolkit,
    reraise_exceptions,
    ToolkitName,
)


class DateModel(HasTraits):

    selected_date = Instance(datetime.date)
    special_days = Dict(Str, List(Instance(datetime.date)))
    styles_mapping = Dict(Str, Instance(CellFormat))


def get_example_model():
    return DateModel(
        special_days={
            "public-holidays": [datetime.date(2020, 1, 1)],
            "weekends": [datetime.date(2020, 1, 12)],
        },
        styles_mapping={
            "public-holidays": CellFormat(bgcolor=(255, 0, 0)),
            "weekends": CellFormat(bold=True),
        },
    )


# StyledDateEditor is currently only implemented for Qt
@requires_toolkit([ToolkitName.qt])
class TestStyledDateEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_init_and_dispose(self):
        # Smoke test to test init and dispose.
        instance = get_example_model()
        view = View(
            Item(
                "selected_date",
                editor=StyledDateEditor(
                    dates_trait="special_days",
                    styles_trait="styles_mapping",
                ),
                style="custom",
            )
        )
        with reraise_exceptions(), \
                create_ui(instance, dict(view=view)):
            pass



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Define the registry object responsible for collecting and reporting
implementations for testing various GUI elements.
"""

import inspect


from traitsui.testing.tester._abstract_target_registry import (
    AbstractTargetRegistry,
)
from traitsui.testing.tester.exceptions import (
    InteractionNotSupported,
    LocationNotSupported,
)


class _TargetToKeyRegistry:
    """ Perform the mapping from target to a key to a callable.

    Internally this is a dict(type, dict(type, callable)), but expose a few
    methods for better error reporting.
    """

    def __init__(self, exception_maker):
        """ Initializer

        Parameters
        ----------
        exception_maker : callable(target_class, key, available_keys)
            A callable that return an exception for when no values are referred
            to by a given pair of target_class and key.
        """
        self._target_to_key_to_value = {}
        self.exception_maker = exception_maker

    def register(self, target_class, key, value):
        action_to_handler = self._target_to_key_to_value.setdefault(
            target_class, {}
        )
        if key in action_to_handler:
            raise ValueError(
                "A value for target {!r} and key {!r} already "
                "exists.".format(target_class, key)
            )
        action_to_handler[key] = value

    def get_value(self, target_class, key):
        action_to_handler = self._target_to_key_to_value.get(target_class, [])
        if key not in action_to_handler:
            raise self.exception_maker(
                target_class=target_class,
                key=key,
                available_keys=list(action_to_handler),
            )
        return action_to_handler[key]

    def get_keys(self, target_class):
        """ Return all the keys for the given target.

        Parameters
        ----------
        target_class : subclass of type
            The type of a UI target being operated on.

        Returns
        -------
        keys : set
        """
        return set(self._target_to_key_to_value.get(target_class, []))


class TargetRegistry(AbstractTargetRegistry):
    """ An object for registering interaction and location resolution logic
    for different UI target types.

    ``register_interaction`` supports extending ``UIWrapper.perform`` and
    ``UIWrapper.inspect`` for a given UI target type and interaction type.

    ``register_location`` supports extending ``UIWrapper.locate`` for a given
    UI target type and location type.

    See :ref:`testing-how-extension-works` in the User Manual for further
    details.
    """

    def __init__(self):
        self._interaction_registry = _TargetToKeyRegistry(
            exception_maker=(
                lambda target_class, key, available_keys: (
                    InteractionNotSupported(
                        target_class=target_class,
                        interaction_class=key,
                        supported=available_keys,
                    )
                )
            ),
        )
        self._location_registry = _TargetToKeyRegistry(
            exception_maker=(
                lambda target_class, key, available_keys: LocationNotSupported(
                    target_class=target_class,
                    locator_class=key,
                    supported=available_keys,
                )
            ),
        )

    def register_interaction(self, target_class, interaction_class, handler):
        """ Register a handler for a given target type and interaction type.

        Parameters
        ----------
        target_class : subclass of type
            The type of a UI target being operated on.
        interaction_class : subclass of type
            Any class.
        handler : callable(UIWrapper, interaction) -> any
            The function to handle the particular interaction on a target.
            ``interaction`` should be an instance of ``interaction_class``.

        Raises
        ------
        ValueError
            If a handler has already be registered for the same target type
            and interaction class.
        """
        self._interaction_registry.register(
            target_class=target_class,
            key=interaction_class,
            value=handler,
        )

    def _get_handler(self, target, interaction):
        """ Return a callable for handling an interaction for a given target.

        Parameters
        ----------
        target : any
            The UI target being operated on.
        interaction : any
            Any interaction object.

        Returns
        -------
        handler : callable(UIWrapper, interaction) -> any
            The function to handle the given interaction on a target.

        Raises
        ------
        InteractionNotSupported
            If the given target and interaction types are not supported
            by this registry.
        """
        return self._interaction_registry.get_value(
            target_class=target.__class__,
            key=interaction.__class__,
        )

    def _get_interactions(self, target):
        """ Returns all the interactions supported for the given target.

        Parameters
        ----------
        target : any
            The UI target for which supported interactions are queried.

        Returns
        -------
        interaction_classes : set
            Supported interaction types for the given target type.
        """
        return self._interaction_registry.get_keys(
            target_class=target.__class__
        )

    def _get_interaction_doc(self, target, interaction_class):
        """ Return the documentation for the given target and interaction type.

        Parameters
        ----------
        target : any
            The UI target for which the interaction will be applied.
        interaction_class : subclass of type
            Any class.

        Returns
        -------
        doc : str

        Raises
        ------
        InteractionNotSupported
            If the given target and interaction types are not supported
            by this registry.
        """
        self._interaction_registry.get_value(
            target_class=target.__class__,
            key=interaction_class,
        )
        # This maybe configurable in the future via register_interaction
        return inspect.getdoc(interaction_class)

    def register_location(self, target_class, locator_class, solver):
        """ Register a solver for resolving the next UI target for the given
        target type and locator type.

        Parameters
        ----------
        target_class : subclass of type
            The type of a UI target being operated on.
        locator_class : subclass of type
            Any class.
        solver : callable(UIWrapper, location) -> any
            A callable for resolving a location into a new target.
            The location argument will be an instance of locator_class.

        Raises
        ------
        ValueError
            If a solver has already been registered for the given target
            type and locator type.
        """
        self._location_registry.register(
            target_class=target_class,
            key=locator_class,
            value=solver,
        )

    def _get_solver(self, target, location):
        """ Return a callable registered for resolving a location for the
        given target and location.

        Parameters
        ----------
        target : any
            The UI target being operated on.
        location : subclass of type
            The location to be resolved on the target.

        Raises
        ------
        LocationNotSupported
            If the given locator and target types are not supported.
        """
        return self._location_registry.get_value(
            target_class=target.__class__,
            key=location.__class__,
        )

    def _get_locations(self, target):
        """ Returns all the location types supported for the given target.

        Parameters
        ----------
        target : any
            The UI target for which supported location types are queried.

        Returns
        -------
        locators_classes : set
            Supported locator types for the given target type.
        """
        return self._location_registry.get_keys(target_class=target.__class__)

    def _get_location_doc(self, target, locator_class):
        """ Return the documentation for the given target and locator type.

        Parameters
        ----------
        target : any
            The UI target being operated on.
        locator_class : subclass of type
            Any class.

        Returns
        -------
        doc : str

        Raises
        ------
        LocationNotSupported
            If the given locator and target types are not supported.
        """
        self._location_registry.get_value(
            target_class=target.__class__,
            key=locator_class,
        )
        # This maybe configurable in the future via register_location
        return inspect.getdoc(locator_class)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Run this script to regenerate the screenshots for the various editors used
on the "The Predefined Trait Editor Factories" and "Advanced Trait Editors"
pages of the documentation.

Note: This script assumes it is being run from the traitsui root directory.
ie. one should run 
`python docs/source/traitsui_user_manual/scripts/regenerate_example_screenshots.py`
"""
import contextlib
import os
from unittest import mock
from functools import partialmethod

import pkg_resources

from traits.api import HasTraits
from traitsui.testing.api import UITester

# Demo files are part of the package data.
DEMO = pkg_resources.resource_filename("traitsui", "examples/demo")


def _is_python_file(path):
    """ Return true if the given path is (public) non-test Python file."""
    _, basename = os.path.split(path)
    _, ext = os.path.splitext(basename)
    return (
        ext == ".py"
        and not basename.startswith("_")
        and not basename.startswith("test_")
    )


def get_python_files(directory):
    """ Report Python files to be run or to be skipped.

    Returns
    -------
    accepted_files : list of str
        Python file paths to be run and screenshotted.
    """
    accepted_files = []
    for root, _, files in os.walk(directory):
        for filename in files:
            path = os.path.abspath(os.path.join(root, filename))
            if not _is_python_file(path):
                continue
            accepted_files.append(path)

    # Avoid arbitrary ordering from the OS
    return sorted(accepted_files)


def get_editor_example_filees(python_files):
    Editor_demo_files = []
    for filename in python_files:
        if filename.endswith('Editor_demo.py'):
            Editor_demo_files.append(filename)
    return Editor_demo_files


def replaced_configure_traits(
    instance,
    filename=None,
    view=None,
    kind=None,
    edit=True,
    context=None,
    handler=None,
    id="",
    scrollable=None,
    screenshot_name=None,
    **args,
):
    """ Mocked configure_traits to launch then close the GUI.
    """
    ui_kwargs = dict(
        view=view,
        parent=None,
        kind="live",
        context=context,
        handler=handler,
        id=id,
        scrollable=scrollable,
        **args,
    )
    with UITester().create_ui(instance, ui_kwargs) as ui:
        ui_pixmap = ui.control.grab()
        ui_pixmap.save(
            os.path.join(
                "docs",
                "source",
                "traitsui_user_manual",
                "images",
                screenshot_name
            )
        )


@contextlib.contextmanager
def replace_configure_traits(screenshot_name):
    """ Context manager to temporarily replace HasTraits.configure_traits
    with a mocked version such that GUI launched are closed soon after they
    are open.
    """
    original_func = HasTraits.configure_traits
    HasTraits.configure_traits = partialmethod(
        replaced_configure_traits, screenshot_name=screenshot_name
    )
    try:
        yield
    finally:
        HasTraits.configure_traits = original_func


def run_file(file_path):
    """ Execute a given Python file.

    Parameters
    ----------
    file_path : str
        File path to be run to retrieve screenshot.
    """
    demo_name = os.path.basename(file_path)
    screenshot_name = demo_name.split('.')[0] + '.png'
    with open(file_path, "r", encoding="utf-8") as f:
        content = f.read()

    globals = {
        "__name__": "__main__",
        "__file__": file_path,
    }
    with replace_configure_traits(screenshot_name), \
            mock.patch("sys.argv", [file_path]):
        # Mock argv: Some example reads sys.argv to allow more arguments
        # But all examples should support being run without additional
        # arguments.
        exec(content, globals)


def main():
    py_files = get_python_files(DEMO)
    example_files = get_editor_example_filees(py_files)

    for filename in example_files:
        run_file(filename)


if __name__ == '__main__':
    main()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import warnings

from pyface.qt import QtCore, QtGui
from pyface.qt.QtTest import QTest

from traitsui.testing.tester._ui_tester_registry._compat import (
    check_key_compat
)
from traitsui.testing.tester.exceptions import Disabled
from traitsui.qt4.key_event_to_name import key_map as _KEY_MAP


def key_click(widget, key, delay):
    """ Performs a key click of the given key on the given widget after
    a delay.

    Parameters
    ----------
    widget : Qwidget
        The Qt widget to be key clicked.
    key : str
        Standardized (pyface) name for a keyboard event.
        e.g. "Enter", "Tab", "Space", "0", "1", "A", ...
        Note: modifiers (e.g. Shift, Alt, etc. are not currently supported)
    delay : int
        Time delay (in ms) in which the key click will be performed.
    """

    mapping = {name: event for event, name in _KEY_MAP.items()}
    if key not in mapping:
        raise ValueError(
            "Unknown key {!r}. Expected one of these: {!r}".format(
                key, sorted(mapping)
            ))
    QTest.keyClick(
        widget,
        mapping[key],
        QtCore.Qt.NoModifier,
        delay=delay,
    )


def check_q_model_index_valid(index):
    """ Checks if a given QModelIndex is valid.

    Parameters
    ----------
    index : QModelIndex

    Raises
    ------
    LookupError
        If the index is not valid.
    """
    if not index.isValid():
        row = index.row()
        column = index.column()
        raise LookupError(
            "Unabled to locate item with row {!r} and column {!r}.".format(
                row, column,
            )
        )


# Generic Handlers ###########################################################


def displayed_text_qobject(widget):
    ''' Helper function to define handlers for various Qwidgets to handle
    query.DisplayedText interactions.

    Parameters
    ----------
    widget : Qwidget
        The Qwidget object with text to be displayed.  Should be one of the
        following QWidgets: 1) QtGui.QLineEdit 2) QtGui.QTextEdit or
        3) QtGui.QLabel

    Notes
    -----
    Qt SimpleEditors occassionally use QtGui.QTextEdit as their control, and
    other times use QtGui.QLineEdit
    '''
    if isinstance(widget, QtGui.QLineEdit):
        return widget.displayText()
    elif isinstance(widget, QtGui.QTextEdit):
        return widget.toPlainText()
    else:
        return widget.text()


def mouse_click_qwidget(control, delay):
    """ Performs a mouce click on a Qt widget.

    Parameters
    ----------
    control : Qwidget
        The Qt widget to be clicked.
    delay : int
        Time delay (in ms) in which click will be performed.
    """

    # for QAbstractButtons we do not use QTest.mouseClick as it assumes the
    # center of the widget as the location to be clicked, which may be
    # incorrect. For QAbstractButtons we can simply call their click method.
    if isinstance(control, QtGui.QAbstractButton):
        if delay > 0:
            QTest.qSleep(delay)
        control.click()
    else:
        QTest.mouseClick(
            control,
            QtCore.Qt.LeftButton,
            delay=delay,
        )


def mouse_click_tab_index(tab_widget, index, delay):
    """ Performs a mouse click on a tab at an index in a QtGui.QTabWidget.

    Parameters
    ----------
    tab_widget : QtGui.QTabWidget
        The tab widget containing the tab to be clicked.
    index : int
        The index of the tab to be clicked.
    delay : int
        Time delay (in ms) in which click will be performed.

    Raises
    ------
    IndexError
        If the index is out of range.
    """
    if not 0 <= index < tab_widget.count():
        raise IndexError(index)
    tabbar = tab_widget.tabBar()
    rect = tabbar.tabRect(index)
    QTest.mouseClick(
        tabbar,
        QtCore.Qt.LeftButton,
        QtCore.Qt.NoModifier,
        rect.center(),
        delay=delay,
    )


def mouse_click_qlayout(layout, index, delay):
    """ Performs a mouse click on a widget at an index in a QLayout.

    Parameters
    ----------
    layout : Qlayout
        The layout containing the widget to be clicked
    index : int
        The index of the widget in the layout to be clicked

    Raises
    ------
    IndexError
        If the index is out of range.
    """
    if not 0 <= index < layout.count():
        raise IndexError(index)
    widget = layout.itemAt(index).widget()
    mouse_click_qwidget(widget, delay)


def mouse_click_item_view(model, view, index, delay):
    """ Perform mouse click on the given QAbstractItemModel (model) and
    QAbstractItemView (view) with the given row and column.

    Parameters
    ----------
    model : QAbstractItemModel
        Model from which QModelIndex will be obtained
    view : QAbstractItemView
        View from which the widget identified by the index will be
        found and key sequence be performed.
    index : QModelIndex

    Raises
    ------
    LookupError
        If the index cannot be located.
        Note that the index error provides more
    """
    check_q_model_index_valid(index)
    rect = view.visualRect(index)
    QTest.mouseClick(
        view.viewport(),
        QtCore.Qt.LeftButton,
        QtCore.Qt.NoModifier,
        rect.center(),
        delay=delay,
    )


def mouse_click_combobox(combobox, index, delay):
    """ Perform a mouse click on a QComboBox at a given index.

    Paramters
    ---------
    combobox : QtGui.ComboBox
        The combobox to be clicked.
    index : int
        The index of the item in the combobox to be clicked.
    delay : int
        Time delay (in ms) in which each key click in the sequence will be
        performed.
    """
    if combobox:
        q_model_index = combobox.model().index(index, 0)
        check_q_model_index_valid(q_model_index)
        mouse_click_item_view(
            model=combobox.model(),
            view=combobox.view(),
            index=q_model_index,
            delay=delay,
        )
        # Otherwise the click won't get registered.
        key_click(
            combobox.view().viewport(), key="Enter", delay=delay
        )
    else:
        warnings.warn(
            "Attempted to click on a non-existant combobox. Nothing was "
            "performed."
        )


def key_sequence_qwidget(control, interaction, delay):
    """ Performs simulated typing of a sequence of keys on the given widget
    after a delay.

    Parameters
    ----------
    control : Qwidget
        The Qt widget to be acted on.
    interaction : instance of command.KeySequence
        The interaction object holding the sequence of key inputs
        to be simulated being typed
    delay : int
        Time delay (in ms) in which each key click in the sequence will be
        performed.

    Raises
    ------
    Disabled
        If the widget is not enabled.
    """
    if not control.isEnabled():
        raise Disabled("{!r} is disabled.".format(control))
    QTest.keyClicks(control, interaction.sequence, delay=delay)


def key_sequence_textbox(control, interaction, delay):
    """ Performs simulated typing of a sequence of keys on a widget that is
    a textbox. The keys are restricted to values also supported for testing
    wx.TextCtrl.

    Parameters
    ----------
    control : QWidget
        The Qt widget intended to hold text for editing.
        e.g. QLineEdit and QTextEdit
    interaction : instance of command.KeySequence
        The interaction object holding the sequence of key inputs
        to be simulated being typed
    delay : int
        Time delay (in ms) in which each key click in the sequence will be
        performed.
    """
    for key in interaction.sequence:
        check_key_compat(key)
    if not control.hasFocus():
        key_click(widget=control, key="End", delay=0)
    key_sequence_qwidget(control=control, interaction=interaction, delay=delay)


def key_click_qwidget(control, interaction, delay):
    """ Performs simulated typing of a key on the given widget after a delay.

    Parameters
    ----------
    control : Qwidget
        The Qt widget to be acted on.
    interaction : instance of command.KeyClick
        The interaction object holding the key input
        to be simulated being typed
    delay : int
        Time delay (in ms) in which the key click will be performed.

    Raises
    ------
    Disabled
        If the widget is not enabled.
    """
    if not control.isEnabled():
        raise Disabled("{!r} is disabled.".format(control))
    key_click(control, interaction.key, delay=delay)


def key_click_qslider(control, interaction, delay):
    """ Performs simulated typing of a key on the given slider after a delay.
    Only allowed keys are:
    "Left", "Right", "Up", "Down", "Page Up", "Page Down"
    Also, note that up related keys correspond to an increment on the slider,
    and down a decrement.

    Parameters
    ----------
    control : QSlider
        The Qt slider to be acted on.
    interaction : instance of command.KeyClick
        The interaction object holding the key input
        to be simulated being typed
    delay : int
        Time delay (in ms) in which the key click will be performed.

    Raises
    ------
    ValueError
        If the interaction.key is not one of the valid keys.
    """
    valid_keys = {"Left", "Right", "Up", "Down", "Page Up", "Page Down"}
    if interaction.key not in valid_keys:
        raise ValueError(
            "Unexpected Key. Supported keys are: {}".format(sorted(valid_keys))
        )
    else:
        key_click(control, interaction.key, delay)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the tree editor factory for all traits user interface toolkits.
"""

from traits.api import Any, Dict, Bool, Tuple, Int, List, Instance, Str, Enum

from traitsui.dock_window_theme import DockWindowTheme
from traitsui.editor_factory import EditorFactory
from traitsui.helper import Orientation
from traitsui.menu import Action
from traitsui.tree_node import TreeNode

# -------------------------------------------------------------------------
#  Trait definitions:
# -------------------------------------------------------------------------

#: Size of each tree node icon
IconSize = Tuple((16, 16), Int, Int)


# -------------------------------------------------------------------------
#  The core tree node menu actions:
# -------------------------------------------------------------------------


NewAction = "NewAction"
CopyAction = Action(
    name="Copy",
    action="editor._menu_copy_node",
    enabled_when="editor._is_copyable(object)",
)
CutAction = Action(
    name="Cut",
    action="editor._menu_cut_node",
    enabled_when="editor._is_cutable(object)",
)
PasteAction = Action(
    name="Paste",
    action="editor._menu_paste_node",
    enabled_when="editor._is_pasteable(object)",
)
DeleteAction = Action(
    name="Delete",
    action="editor._menu_delete_node",
    enabled_when="editor._is_deletable(object)",
)
RenameAction = Action(
    name="Rename",
    action="editor._menu_rename_node",
    enabled_when="editor._is_renameable(object)",
)


class TreeEditor(EditorFactory):
    """ Editor factory for tree editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Supported TreeNode objects
    nodes = List(TreeNode)

    #: Mapping from TreeNode tuples to MultiTreeNodes
    multi_nodes = Dict()

    #: The column header labels if any.
    column_headers = List(Str)

    #: Are the individual nodes editable?
    editable = Bool(True)

    #: Selection mode.
    selection_mode = Enum("single", "extended")

    #: Is the editor shared across trees?
    shared_editor = Bool(False)

    #: Reference to a shared object editor
    editor = Instance(EditorFactory)

    # FIXME: Implemented only in wx backend.
    #: The DockWindow graphical theme
    dock_theme = Instance(DockWindowTheme)

    #: Show icons for tree nodes?
    show_icons = Bool(True)

    #: Hide the tree root node?
    hide_root = Bool(False)

    #: Layout orientation of the tree and the editor
    orientation = Orientation

    #: Number of tree levels (down from the root) that should be automatically
    #: opened
    auto_open = Int()

    #: Size of the tree node icons
    icon_size = IconSize

    #: Called when a node is selected
    on_select = Any()

    #: Called when a node is clicked
    on_click = Any()

    #: Called when a node is double-clicked
    on_dclick = Any()

    #: Called when a node is activated
    on_activated = Any()

    #: Call when the mouse hovers over a node
    on_hover = Any()

    #: The optional extended trait name of the trait to synchronize with the
    #: editor's current selection:
    selected = Str()

    #: The optional extended trait name of the trait that should be assigned
    #: a node object when a tree node is activated, by double-clicking or
    #: pressing the Enter key when a node has focus (Note: if you want to
    #: receive repeated activated events on the same node, make sure the trait
    #: is defined as an Event):
    activated = Str()

    #: The optional extended trait name of the trait that should be assigned
    #: a node object when a tree node is clicked on (Note: If you want to
    #: receive repeated clicks on the same node, make sure the trait is defined
    #: as an Event):
    click = Str()

    #: The optional extended trait name of the trait that should be assigned
    #: a node object when a tree node is double-clicked on (Note: if you want
    #: to receive repeated double-clicks on the same node, make sure the trait
    #: is defined as an Event):
    dclick = Str()

    #: The optional extended trait name of the trait event that is fired
    #: whenever the application wishes to veto a tree action in progress (e.g.
    #: double-clicking a non-leaf tree node normally opens or closes the node,
    #: but if you are handling the double-click event in your program, you may
    #: wish to veto the open or close operation). Be sure to fire the veto
    #: event in the event handler triggered by the operation (e.g. the 'dclick'
    #: event handler.
    veto = Str()

    #: The optional extended trait name of the trait event that is fired when
    #: the application wishes the currently visible portion of the tree widget
    #: to repaint itself.
    refresh = Str()

    #: Mode for lines connecting tree nodes
    #:
    #: * 'appearance': Show lines only when they look good.
    #: * 'on': Always show lines.
    #: * 'off': Don't show lines.
    lines_mode = Enum("appearance", "on", "off")

    # FIXME: Document as unimplemented or wx specific.
    #: Whether to alternate row colors or not.
    alternating_row_colors = Bool(False)

    #: Any extra vertical padding to add.
    vertical_padding = Int(0)

    #: Whether or not to expand on a double-click.
    expands_on_dclick = Bool(True)

    #: Whether the labels should be wrapped around, if not an ellipsis is shown
    #: This works only in the qt backend and if there is only one column in
    #: tree
    word_wrap = Bool(False)


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = TreeEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
A demo to demonstrate the progress column in a TableEditor. The demo only works
with qt backend.
"""

import random

from pyface.api import GUI
from traits.api import Button, HasTraits, Instance, Int, List, Str
from traitsui.api import ObjectColumn, TableEditor, UItem, View
from traitsui.extras.progress_column import ProgressColumn


class Job(HasTraits):

    name = Str()

    percent_complete = Int()


class JobManager(HasTraits):

    jobs = List(Instance(Job))

    start = Button()

    def populate(self):
        self.jobs = [
            Job(name='job %02d' % i, percent_complete=0)
            for i in range(1, 25)
        ]

    def process(self):
        for job in self.jobs:
            job.percent_complete = min(
                job.percent_complete + random.randint(0, 3), 100)

        if any(job.percent_complete < 100 for job in self.jobs):
            GUI.invoke_after(100, self.process)

    def _start_fired(self):
        self.populate()
        GUI.invoke_after(1000, self.process)

    traits_view = View(
        UItem(
            'jobs',
            editor=TableEditor(
                columns=[
                    ObjectColumn(name='name'),
                    ProgressColumn(name='percent_complete'),
                ]
            )
        ),
        UItem('start'),
        resizable=True,
    )


demo = JobManager()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traitsui.testing.tester.target_registry import TargetRegistry
from traitsui.testing.tester._ui_tester_registry.qt4._traitsui import (
    boolean_editor,
    button_editor,
    check_list_editor,
    editor_factory,
    enum_editor,
    file_editor,
    font_editor,
    instance_editor,
    list_editor,
    range_editor,
    text_editor,
    ui_base,
)
from ._control_widget_registry import get_widget_registry


def get_default_registries():
    """ Creates the default registries for UITester that are qt specific.

    Returns
    -------
    registries : list of AbstractTargetRegistry
        The default registries containing implementations for TraitsUI editors
        that are qt specific.
    """
    registry = TargetRegistry()

    # BooleanEditor
    boolean_editor.register(registry)

    # ButtonEditor
    button_editor.register(registry)

    # CheckListEditor
    check_list_editor.register(registry)

    # EnumEditor
    enum_editor.register(registry)

    # FileEditor
    file_editor.register(registry)

    # FontEditor
    font_editor.register(registry)

    # TextEditor
    text_editor.register(registry)

    # ListEditor
    list_editor.register(registry)

    # RangeEditor
    range_editor.register(registry)

    # ui_base
    ui_base.register(registry)

    # InstanceEditor
    instance_editor.register(registry)

    # Editor Factory
    editor_factory.register(registry)

    # The more general registry goes after the more specific registry.
    return [
        registry,
        get_widget_registry(),
    ]



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest import mock

from traitsui.tests._tools import (
    is_qt,
    requires_toolkit,
    ToolkitName,
)
from traitsui.testing.tester import command
from traitsui.testing.tester.exceptions import Disabled
from traitsui.testing.tester._ui_tester_registry.qt4 import (
    _interaction_helpers
)

try:
    from pyface.qt import QtGui
except ImportError:
    if is_qt():
        raise


@requires_toolkit([ToolkitName.qt])
class TestInteractions(unittest.TestCase):

    def test_mouse_click(self):
        button = QtGui.QPushButton()
        click_slot = mock.Mock()
        button.clicked.connect(click_slot)

        _interaction_helpers.mouse_click_qwidget(button, 0)

        self.assertEqual(click_slot.call_count, 1)

    def test_mouse_click_disabled(self):
        button = QtGui.QPushButton()
        button.setEnabled(False)

        click_slot = mock.Mock()
        button.clicked.connect(click_slot)

        # when
        # clicking won't fail, it just does not do anything.
        # This is consistent with the actual UI.
        _interaction_helpers.mouse_click_qwidget(button, 0)

        # then
        self.assertEqual(click_slot.call_count, 0)

    def test_mouse_click_combobox_warns(self):
        combo = None
        with self.assertWarns(UserWarning):
            _interaction_helpers.mouse_click_combobox(combo, 0, 0)

    def test_key_sequence(self):
        # test on different Qwidget objects
        textboxes = [QtGui.QLineEdit(), QtGui.QTextEdit()]
        for i, textbox in enumerate(textboxes):
            change_slot = mock.Mock()
            textbox.textChanged.connect(change_slot)

            # when
            _interaction_helpers.key_sequence_qwidget(
                textbox, command.KeySequence("abc"), 0)

            # then
            if i == 0:
                self.assertEqual(textbox.text(), "abc")
            else:
                self.assertEqual(textbox.toPlainText(), "abc")
            # each keystroke fires a signal
            self.assertEqual(change_slot.call_count, 3)

        # for a QLabel, one can try a key sequence and nothing will happen
        textbox = QtGui.QLabel()
        _interaction_helpers.key_sequence_qwidget(
            textbox, command.KeySequence("abc"), 0)
        self.assertEqual(textbox.text(), "")

    def test_key_sequence_textbox_with_unicode(self):
        for code in range(32, 127):
            with self.subTest(code=code, word=chr(code)):
                textbox = QtGui.QLineEdit()
                change_slot = mock.Mock()
                textbox.textChanged.connect(change_slot)

                # when
                _interaction_helpers.key_sequence_textbox(
                    textbox,
                    command.KeySequence(chr(code) * 3),
                    delay=0,
                )

                # then
                self.assertEqual(textbox.text(), chr(code) * 3)
                self.assertEqual(change_slot.call_count, 3)

    def test_key_sequence_unsupported_key(self):
        textbox = QtGui.QLineEdit()

        with self.assertRaises(ValueError) as exception_context:
            # QTest does not support this character.
            _interaction_helpers.key_sequence_textbox(
                textbox,
                command.KeySequence(chr(31)),
                delay=0,
            )

        self.assertIn(
            "is currently not supported.",
            str(exception_context.exception),
        )

    def test_key_sequence_backspace_character(self):
        # Qt does convert backspace character to the backspace key
        # But we disallow it for now to be consistent with wx.
        textbox = QtGui.QLineEdit()

        with self.assertRaises(ValueError) as exception_context:
            _interaction_helpers.key_sequence_textbox(
                textbox,
                command.KeySequence("\b"),
                delay=0,
            )

        self.assertIn(
            "is currently not supported.",
            str(exception_context.exception),
        )

    def test_key_sequence_insert_point_qlineedit(self):
        textbox = QtGui.QLineEdit()
        textbox.setText("123")

        # when
        _interaction_helpers.key_sequence_textbox(
            textbox,
            command.KeySequence("abc"),
            delay=0,
        )

        # then
        self.assertEqual(textbox.text(), "123abc")

    def test_key_sequence_insert_point_qtextedit(self):
        # The default insertion point moved to the end to be consistent
        # with QLineEdit
        textbox = QtGui.QTextEdit()
        textbox.setText("123")

        # when
        _interaction_helpers.key_sequence_textbox(
            textbox,
            command.KeySequence("abc"),
            delay=0,
        )

        # then
        self.assertEqual(textbox.toPlainText(), "123abc")

    def test_key_sequence_disabled(self):
        textbox = QtGui.QLineEdit()
        textbox.setEnabled(False)

        # this will fail, because one should not be allowed to set
        # cursor on the widget to type anything
        with self.assertRaises(Disabled):
            _interaction_helpers.key_sequence_qwidget(
                textbox, command.KeySequence("abc"), 0)

    def test_key_click(self):
        textbox = QtGui.QLineEdit()
        change_slot = mock.Mock()
        textbox.editingFinished.connect(change_slot)

        # sanity check on editingFinished signal
        _interaction_helpers.key_sequence_qwidget(
            textbox, command.KeySequence("abc"), 0)
        self.assertEqual(change_slot.call_count, 0)

        _interaction_helpers.key_click_qwidget(
            textbox, command.KeyClick("Enter"), 0)
        self.assertEqual(change_slot.call_count, 1)

        # test on a different Qwidget object - QtGui.QTextEdit()
        textbox = QtGui.QTextEdit()
        change_slot = mock.Mock()
        # Now "Enter" should not finish editing, but instead go to next line
        textbox.textChanged.connect(change_slot)
        _interaction_helpers.key_click_qwidget(
            textbox, command.KeyClick("Enter"), 0)
        self.assertEqual(change_slot.call_count, 1)
        self.assertEqual(textbox.toPlainText(), "\n")

        # for a QLabel, one can try a key click and nothing will happen
        textbox = QtGui.QLabel()
        _interaction_helpers.key_click_qwidget(
            textbox, command.KeyClick("A"), 0)
        self.assertEqual(textbox.text(), "")

    def test_key_click_disabled(self):
        textbox = QtGui.QLineEdit()
        textbox.setEnabled(False)
        change_slot = mock.Mock()
        textbox.editingFinished.connect(change_slot)

        with self.assertRaises(Disabled):
            _interaction_helpers.key_click_qwidget(
                textbox, command.KeyClick("Enter"), 0)
        self.assertEqual(change_slot.call_count, 0)

    def test_check_q_model_index_valid(self):
        self.widget = QtGui.QListWidget()
        self.items = ["a", "b", "c"]
        self.widget.addItems(self.items)
        self.good_q_index = self.widget.model().index(1, 0)
        self.bad_q_index = self.widget.model().index(10, 0)

        self.model = self.widget.model()
        _interaction_helpers.check_q_model_index_valid(self.good_q_index)
        with self.assertRaises(LookupError):
            _interaction_helpers.check_q_model_index_valid(self.bad_q_index)

    def test_key_click_q_slider_helpful_err(self):
        slider = QtGui.QSlider()
        with self.assertRaises(ValueError) as exc:
            _interaction_helpers.key_click_qslider(
                slider, command.KeyClick("Enter"), 0)
        self.assertIn(
            "['Down', 'Left', 'Page Down', 'Page Up', 'Right', 'Up']",
            str(exc.exception)
        )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

Another demo showing how to use a TabularEditor to create a multi-select list
box. This demo creates a reusable StringListEditor class and uses that instead
of defining the editor as part of the demo class.

This approach greatly simplifies the actual demo class and shows how to
construct a reusable Traits UI-based editor that can be used in other
applications.
"""
# Issue related to the demo warning: enthought/traitsui#960

from traits.api import HasPrivateTraits, List, Str, Property, observe
from traits.etsconfig.api import ETSConfig

from traitsui.api import (
    BasicEditorFactory, HGroup, Item, TabularAdapter, TabularEditor, View
)

if ETSConfig.toolkit == 'wx':
    from traitsui.wx.ui_editor import UIEditor
else:
    from traitsui.qt4.ui_editor import UIEditor


# -- Define the reusable StringListEditor class and its helper classes --------

# Define the tabular adapter used by the Traits UI string list editor:
class MultiSelectAdapter(TabularAdapter):

    # The columns in the table (just the string value):
    columns = [('Value', 'value')]

    # The text property used for the 'value' column:
    value_text = Property()

    def _get_value_text(self):
        return self.item


# Define the actual Traits UI string list editor:
class _StringListEditor(UIEditor):

    # Indicate that the editor is scrollable/resizable:
    scrollable = True

    # The list of available editor choices:
    choices = List(Str)

    # The list of currently selected items:
    selected = List(Str)

    # The traits UI view used by the editor:
    traits_view = View(
        Item(
            'choices',
            show_label=False,
            editor=TabularEditor(
                show_titles=False,
                selected='selected',
                editable=False,
                multi_select=True,
                adapter=MultiSelectAdapter()
            )
        ),
        id='string_list_editor',
        resizable=True
    )

    def init_ui(self, parent):

        self.sync_value(self.factory.choices, 'choices', 'from', is_list=True)
        self.selected = self.value

        return self.edit_traits(parent=parent, kind='subpanel')

    @observe('selected')
    def _selected_modified(self, event):
        self.value = self.selected


# Define the StringListEditor class used by client code:
class StringListEditor(BasicEditorFactory):

    # The editor implementation class:
    klass = _StringListEditor

    # The extended trait name containing the editor's set of choices:
    choices = Str()


# -- Define the demo class ----------------------------------------------------
class MultiSelect(HasPrivateTraits):
    """ This class demonstrates using the StringListEditor to select a set
        of string values from a set of choices.
    """

    # The list of choices to select from:
    choices = List(Str)

    # The currently selected list of choices:
    selected = List(Str)

    # A dummy result so that we can display the selection using the same
    # StringListEditor:
    result = List(Str)

    # A traits view showing the list of choices on the left-hand side, and
    # the currently selected choices on the right-hand side:
    traits_view = View(
        HGroup(
            Item(
                'selected',
                show_label=False,
                editor=StringListEditor(choices='choices')
            ),
            Item(
                'result',
                show_label=False,
                editor=StringListEditor(choices='selected')
            )
        ),
        width=0.20,
        height=0.25
    )


# Create the demo:
demo = MultiSelect(
    choices=[
        'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine',
        'ten'
    ],
    selected=['two', 'five', 'nine']
)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
This example demonstrates how to test interacting with a textbox created
using FileEditor.

The GUI being tested is written in the demo under the same name (minus the
preceding 'test') in the outer directory.
"""

import os
import runpy
import unittest

from traitsui.testing.api import (
    DisplayedText, KeyClick, KeySequence, UITester
)

#: Filename of the demo script
FILENAME = "FileEditor_demo.py"

#: Path of the demo script
DEMO_PATH = os.path.join(os.path.dirname(__file__), "..", FILENAME)


class TestFileEditorDemo(unittest.TestCase):

    def test_run_demo(self):
        demo = runpy.run_path(DEMO_PATH)["demo"]

        tester = UITester()
        with tester.create_ui(demo) as ui:
            # simple FileEditor
            simple_file_field = tester.find_by_id(ui, "simple_file")

            # Modify the value on the GUI
            # The path does not exist, which is allowed by the trait.
            # The custom FileEditor will ignore nonexisting file path.
            simple_file_field.perform(KeySequence("some_path"))
            simple_file_field.perform(KeyClick("Enter"))
            self.assertEqual(demo.file_name, "some_path")

            # Modify the value on the object
            demo.file_name = FILENAME
            self.assertEqual(
                simple_file_field.inspect(DisplayedText()), FILENAME
            )

            # On wx, an assertion error occurs upon disposing the UI.
            # That error is linked to the custom FileEditor.
            # See enthought/traitsui#752


# Run the test(s)
unittest.TextTestRunner().run(
    unittest.TestLoader().loadTestsFromTestCase(TestFileEditorDemo)
)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This module provides functions for registering interaction handlers
and location solvers for common Wx GUI components.
"""

from traitsui.testing.tester.command import (
    KeyClick,
    KeySequence,
    MouseClick
)
from traitsui.testing.tester.query import DisplayedText
from traitsui.testing.tester._ui_tester_registry.wx import _interaction_helpers


def register_editable_textbox_handlers(registry, target_class, widget_getter):
    """ Register common interactions for an editable textbox (in Wx)

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to.
    target_class : subclass of type
        The type of target being wrapped in a UIWrapper on which the
        interaction will be performed.
    widget_getter : callable(wrapper: UIWrapper) -> wx.TextCtrl
        A callable to return a wx.TextCtrl
    """
    handlers = [
        (KeySequence,
            (lambda wrapper, interaction:
                _interaction_helpers.key_sequence_text_ctrl(
                    widget_getter(wrapper), interaction, wrapper.delay))),
        (KeyClick,
            (lambda wrapper, interaction:
                _interaction_helpers.key_click_text_entry(
                    widget_getter(wrapper), interaction, wrapper.delay))),
        (MouseClick,
            (lambda wrapper, _: _interaction_helpers.mouse_click_object(
                control=widget_getter(wrapper), delay=wrapper.delay))),
        (DisplayedText,
            lambda wrapper, _: widget_getter(wrapper).GetValue()),
    ]
    for interaction_class, handler in handlers:
        registry.register_interaction(
            target_class=target_class,
            interaction_class=interaction_class,
            handler=handler,
        )
